import torch
import os
import lightning as pl
from omegaconf import OmegaConf
from common.utils import rank_zero_print, get_class
from common.dataset import AspectRatioDataset, worker_init_fn
from modules.sdxl_model import StableDiffusionModel
from lightning.pytorch.utilities.model_summary import ModelSummary

def setup(fabric: pl.Fabric, config: OmegaConf) -> tuple:
    model_path = config.trainer.model_path
    model = SupervisedFineTune(
        model_path=model_path, 
        config=config, 
        device=fabric.device
    )
    dataset = AspectRatioDataset(
        batch_size=config.trainer.batch_size,
        rank=fabric.global_rank,
        dtype=torch.float32,
        base_len=config.trainer.resolution,
        **config.dataset,
    )
    dataloader = torch.utils.data.DataLoader(
        dataset,
        sampler=None,
        batch_size=None,
        persistent_workers=False,
        num_workers=config.dataset.get("num_workers", 4),
        worker_init_fn=worker_init_fn,
        shuffle=False,
        pin_memory=True,
    )
    
    params_to_optim = [{'params': model.model.parameters()}]
    if config.advanced.get("train_text_encoder_1"):
        lr = config.advanced.get("text_encoder_1_lr", config.optimizer.params.lr)
        params_to_optim.append(
            {"params": model.conditioner.embedders[0].parameters(), "lr": lr}
        )
        
    if config.advanced.get("train_text_encoder_2"):
        lr = config.advanced.get("text_encoder_2_lr", config.optimizer.params.lr)
        params_to_optim.append(
            {"params": model.conditioner.embedders[1].parameters(), "lr": lr}
        )

    optim_param = config.optimizer.params
    optimizer = get_class(config.optimizer.name)(
        params_to_optim, **optim_param
    )
    scheduler = None
    if config.get("scheduler"):
        scheduler = get_class(config.scheduler.name)(
            optimizer, **config.scheduler.params
        )
        
    model.first_stage_model.to(torch.float32)
    if fabric.is_global_zero and os.name != "nt":
        print(f"\n{ModelSummary(model, max_depth=1)}\n")
        
    model, optimizer = fabric.setup(model, optimizer)
    dataloader = fabric.setup_dataloaders(dataloader)
    return model, dataset, dataloader, optimizer, scheduler

class SupervisedFineTune(StableDiffusionModel):
    def encode_batch(self, batch):
        self.conditioner.to(self.target_device)
        return self.conditioner(batch)
    
    def forward(self, batch):
        advanced = self.config.get("advanced", {})
        if not batch["is_latent"]:
            self.first_stage_model.to(self.target_device)
            latents = self.encode_first_stage(batch["pixels"].to(self.first_stage_model.dtype))
            if torch.any(torch.isnan(latents)):
                rank_zero_print("NaN found in latents, replacing with zeros")
                latents = torch.where(torch.isnan(latents), torch.zeros_like(latents), latents)
        else:
            self.first_stage_model.cpu()
            latents = batch["pixels"]

        cond = self.encode_batch(batch)
        model_dtype = next(self.model.parameters()).dtype
        cond = {k: v.to(model_dtype) for k, v in cond.items()}

        # Sample noise that we'll add to the latents
        noise = torch.randn_like(latents, dtype=model_dtype)
        if advanced.get("offset_noise"):
            offset = torch.randn(latents.shape[0], latents.shape[1], 1, 1, device=latents.device)
            noise = torch.randn_like(latents) + float(advanced.get("offset_noise_val")) * offset

        # https://arxiv.org/abs/2301.11706
        if advanced.get("input_perturbation"):
            noise = noise + float(advanced.get("input_perturbation_val")) * torch.randn_like(noise)

        bsz = latents.shape[0]

        # Sample a random timestep for each image
        timesteps = torch.randint(
            advanced.get("timestep_start", 0),
            advanced.get("timestep_end", 1000),
            (bsz,),
            dtype=torch.int64,
            device=latents.device,
        )

        # Add noise to the latents according to the noise magnitude at each timestep
        # (this is the forward diffusion process)
        noisy_latents = self.noise_scheduler.add_noise(latents, noise, timesteps)

        # Predict the noise residual
        noise_pred = self.model(noisy_latents, timesteps, cond)

        # Get the target for loss depending on the prediction type
        is_v = advanced.get("v_parameterization", False)
        target = noise if not is_v else self.noise_scheduler.get_velocity(latents, noise, timesteps)

        loss = torch.nn.functional.mse_loss(noise_pred.float(), target.float(), reduction="mean")
        if torch.isnan(loss).any() or torch.isinf(loss).any():
            raise FloatingPointError("Error infinite or NaN loss detected")

        return loss



import re
from abc import abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, Tuple, Union

import lightning as pl
import torch
from omegaconf import ListConfig
from packaging import version
from safetensors.torch import load_file as load_safetensors

from .vae_model import Decoder, Encoder
from .encoder_util import default, get_obj_from_str, instantiate_from_config

class DiagonalGaussianDistribution(object):
    def __init__(self, parameters, deterministic=False):
        self.parameters = parameters
        self.mean, self.logvar = torch.chunk(parameters, 2, dim=1)
        self.logvar = torch.clamp(self.logvar, -30.0, 20.0)
        self.deterministic = deterministic
        self.std = torch.exp(0.5 * self.logvar)
        self.var = torch.exp(self.logvar)
        if self.deterministic:
            self.var = self.std = torch.zeros_like(self.mean).to(
                device=self.parameters.device
            )

    def sample(self):
        x = self.mean + self.std * torch.randn(self.mean.shape).to(
            device=self.parameters.device
        )
        return x

    def kl(self, other=None):
        if self.deterministic:
            return torch.Tensor([0.0])
        else:
            if other is None:
                return 0.5 * torch.sum(
                    torch.pow(self.mean, 2) + self.var - 1.0 - self.logvar,
                    dim=[1, 2, 3],
                )
            else:
                return 0.5 * torch.sum(
                    torch.pow(self.mean - other.mean, 2) / other.var
                    + self.var / other.var
                    - 1.0
                    - self.logvar
                    + other.logvar,
                    dim=[1, 2, 3],
                )

    def nll(self, sample, dims=[1, 2, 3]):
        if self.deterministic:
            return torch.Tensor([0.0])
        logtwopi = np.log(2.0 * np.pi)
        return 0.5 * torch.sum(
            logtwopi + self.logvar + torch.pow(sample - self.mean, 2) / self.var,
            dim=dims,
        )

    def mode(self):
        return self.mean


class AbstractAutoencoder(pl.LightningModule):
    """
    This is the base class for all autoencoders, including image autoencoders, image autoencoders with discriminators,
    unCLIP models, etc. Hence, it is fairly general, and specific features
    (e.g. discriminator training, encoding, decoding) must be implemented in subclasses.
    """

    def __init__(
        self,
        ema_decay: Union[None, float] = None,
        monitor: Union[None, str] = None,
        input_key: str = "jpg",
        ckpt_path: Union[None, str] = None,
        ignore_keys: Union[Tuple, list, ListConfig] = (),
    ):
        super().__init__()
        self.input_key = input_key
        self.use_ema = ema_decay is not None
        if monitor is not None:
            self.monitor = monitor

        if ckpt_path is not None:
            self.init_from_ckpt(ckpt_path, ignore_keys=ignore_keys)

        if version.parse(torch.__version__) >= version.parse("2.0.0"):
            self.automatic_optimization = False

    def init_from_ckpt(
        self, path: str, ignore_keys: Union[Tuple, list, ListConfig] = tuple()
    ) -> None:
        if path.endswith("ckpt"):
            sd = torch.load(path, map_location="cpu")["state_dict"]
        elif path.endswith("safetensors"):
            sd = load_safetensors(path)
        else:
            raise NotImplementedError

        keys = list(sd.keys())
        for k in keys:
            for ik in ignore_keys:
                if re.match(ik, k):
                    print("Deleting key {} from state_dict.".format(k))
                    del sd[k]
        missing, unexpected = self.load_state_dict(sd, strict=False)
        print(
            f"Restored from {path} with {len(missing)} missing and {len(unexpected)} unexpected keys"
        )
        if len(missing) > 0:
            print(f"Missing Keys: {missing}")
        if len(unexpected) > 0:
            print(f"Unexpected Keys: {unexpected}")

    @abstractmethod
    def get_input(self, batch) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def encode(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("encode()-method of abstract base class called")

    @abstractmethod
    def decode(self, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("decode()-method of abstract base class called")

    def instantiate_optimizer_from_config(self, params, lr, cfg):
        print(f"loading >>> {cfg['target']} <<< optimizer from config")
        return get_obj_from_str(cfg["target"])(
            params, lr=lr, **cfg.get("params", dict())
        )


class AutoencodingEngine(AbstractAutoencoder):
    """
    Base class for all image autoencoders that we train, like VQGAN or AutoencoderKL
    (we also restore them explicitly as special cases for legacy reasons).
    Regularizations such as KL or VQ are moved to the regularizer class.
    """

    def __init__(
        self,
        *args,
        encoder_config: Dict,
        decoder_config: Dict,
        loss_config: Dict,
        regularizer_config: Dict,
        optimizer_config: Union[Dict, None] = None,
        lr_g_factor: float = 1.0,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # todo: add options to freeze encoder/decoder
        self.encoder = instantiate_from_config(encoder_config)
        self.decoder = instantiate_from_config(decoder_config)
        self.loss = instantiate_from_config(loss_config)
        self.regularization = instantiate_from_config(regularizer_config)
        self.optimizer_config = default(
            optimizer_config, {"target": "torch.optim.Adam"}
        )
        self.lr_g_factor = lr_g_factor

    def get_input(self, batch: Dict) -> torch.Tensor:
        # assuming unified data format, dataloader returns a dict.
        # image tensors should be scaled to -1 ... 1 and in channels-first format (e.g., bchw instead if bhwc)
        return batch[self.input_key]

    def get_autoencoder_params(self) -> list:
        params = (
            list(self.encoder.parameters())
            + list(self.decoder.parameters())
            + list(self.regularization.get_trainable_parameters())
            + list(self.loss.get_trainable_autoencoder_parameters())
        )
        return params

    def get_discriminator_params(self) -> list:
        params = list(self.loss.get_trainable_parameters())  # e.g., discriminator
        return params

    def get_last_layer(self):
        return self.decoder.get_last_layer()

    def encode(self, x: Any, return_reg_log: bool = False) -> Any:
        z = self.encoder(x)
        z, reg_log = self.regularization(z)
        if return_reg_log:
            return z, reg_log
        return z

    def decode(self, z: Any) -> torch.Tensor:
        x = self.decoder(z)
        return x

    def forward(self, x: Any) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        z, reg_log = self.encode(x, return_reg_log=True)
        dec = self.decode(z)
        return z, dec, reg_log

    @torch.no_grad()
    def log_images(self, batch: Dict, **kwargs) -> Dict:
        log = dict()
        x = self.get_input(batch)
        _, xrec, _ = self(x)
        log["inputs"] = x
        log["reconstructions"] = xrec
        with self.ema_scope():
            _, xrec_ema, _ = self(x)
            log["reconstructions_ema"] = xrec_ema
        return log


class AutoencoderKL(AutoencodingEngine):
    def __init__(self, embed_dim: int, **kwargs):
        ddconfig = kwargs.pop("ddconfig")
        ckpt_path = kwargs.pop("ckpt_path", None)
        ignore_keys = kwargs.pop("ignore_keys", ())
        super().__init__(
            encoder_config={"target": "torch.nn.Identity"},
            decoder_config={"target": "torch.nn.Identity"},
            regularizer_config={"target": "torch.nn.Identity"},
            loss_config=kwargs.pop("lossconfig"),
            **kwargs,
        )
        assert ddconfig["double_z"]
        self.encoder = Encoder(**ddconfig)
        self.decoder = Decoder(**ddconfig)
        self.quant_conv = torch.nn.Conv2d(2 * ddconfig["z_channels"], 2 * embed_dim, 1)
        self.post_quant_conv = torch.nn.Conv2d(embed_dim, ddconfig["z_channels"], 1)
        self.embed_dim = embed_dim

        if ckpt_path is not None:
            self.init_from_ckpt(ckpt_path, ignore_keys=ignore_keys)

    def encode(self, x):
        assert not self.training, f"{self.__class__.__name__} only supports inference currently"
        h = self.encoder(x)
        moments = self.quant_conv(h)
        posterior = DiagonalGaussianDistribution(moments)
        return posterior

    def decode(self, z, **decoder_kwargs):
        z = self.post_quant_conv(z)
        dec = self.decoder(z, **decoder_kwargs)
        return dec


class AutoencoderKLInferenceWrapper(AutoencoderKL):
    def encode(self, x):
        return super().encode(x).sample()


class IdentityFirstStage(AbstractAutoencoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_input(self, x: Any) -> Any:
        return x

    def encode(self, x: Any, *args, **kwargs) -> Any:
        return x

    def decode(self, x: Any, *args, **kwargs) -> Any:
        return x



import torch
import os
import lightning as pl
from omegaconf import OmegaConf
from common.utils import load_torch_file, get_class, EmptyInitWrapper, rank_zero_print
from common.dataset import AspectRatioDataset, worker_init_fn
from modules.train import SupervisedFineTune
from modules.sdxl_utils import get_hidden_states_sdxl, convert_sdxl_text_encoder_2_checkpoint
from modules.sdxl_utils import get_size_embeddings
from lightning.pytorch.utilities.model_summary import ModelSummary
from safetensors.torch import save_file

from diffusers import DDPMScheduler
from transformers import CLIPTokenizer, CLIPTextConfig, CLIPTextModel, CLIPTextModelWithProjection

try:
    from lycoris import create_lycoris, LycorisNetwork
except ImportError as e:
    raise ImportError(f"\n\nError import lycoris: {e} \nTry install lycoris using `pip install lycoris_lora toml`")

def setup(fabric: pl.Fabric, config: OmegaConf) -> tuple:
    model_path = config.trainer.model_path
    model = StableDiffusionModel(
        model_path=model_path, 
        config=config, 
        device=fabric.device
    )
    dataset = AspectRatioDataset(
        batch_size=config.trainer.batch_size,
        rank=fabric.global_rank,
        dtype=torch.float32,
        base_len=config.trainer.resolution,
        **config.dataset,
    )
    dataloader = torch.utils.data.DataLoader(
        dataset,
        sampler=None,
        batch_size=None,
        persistent_workers=False,
        num_workers=config.dataset.get("num_workers", 4),
        worker_init_fn=worker_init_fn,
        shuffle=False,
        pin_memory=True,
    )
    
    params_to_optim = [{"params": model.lycoris_unet.parameters()}]
    # params_to_optim = [{'params': model.model.parameters()}]
    if config.advanced.get("train_text_encoder_1"):
        lr = config.advanced.get("text_encoder_1_lr", config.optimizer.params.lr)
        params_to_optim.append({"params": model.lycoris_te1.parameters(), "lr": lr})
            
    if config.advanced.get("train_text_encoder_2"):
        lr = config.advanced.get("text_encoder_2_lr", config.optimizer.params.lr)
        params_to_optim.append({"params": model.lycoris_te2.parameters(), "lr": lr})

    optim_param = config.optimizer.params
    optimizer = get_class(config.optimizer.name)(
        params_to_optim, **optim_param
    )
    scheduler = None
    if config.get("scheduler"):
        scheduler = get_class(config.scheduler.name)(
            optimizer, **config.scheduler.params
        )
        
    model.first_stage_model.to(torch.float32)
    if fabric.is_global_zero and os.name != "nt":
        print(f"\n{ModelSummary(model, max_depth=1)}\n")
        
    model, optimizer = fabric.setup(model, optimizer)
    dataloader = fabric.setup_dataloaders(dataloader)
    return model, dataset, dataloader, optimizer, scheduler


def init_text_encoder():
    text_model_1_cfg = CLIPTextConfig.from_pretrained("stabilityai/stable-diffusion-xl-base-1.0", subfolder="text_encoder")
    text_model_2_cfg = CLIPTextConfig.from_pretrained("stabilityai/stable-diffusion-xl-base-1.0", subfolder="text_encoder_2")
    
    with EmptyInitWrapper():    
        text_model_1 = CLIPTextModel._from_config(text_model_1_cfg)
        text_model_2 = CLIPTextModelWithProjection._from_config(text_model_2_cfg)
        
    tokenizer1 = CLIPTokenizer.from_pretrained("stabilityai/stable-diffusion-xl-base-1.0", subfolder="tokenizer")
    tokenizer2 = CLIPTokenizer.from_pretrained("stabilityai/stable-diffusion-xl-base-1.0", subfolder="tokenizer_2")
    text_model_1 = text_model_1.requires_grad_(False)
    text_model_2 = text_model_2.requires_grad_(False)
    return text_model_1, text_model_2, tokenizer1, tokenizer2


# define the LightningModule
class StableDiffusionModel(SupervisedFineTune):
    def init_model(self):
        trainer_cfg = self.config.trainer
        sd = load_torch_file(self.model_path, self.target_device)
        vae, unet, _ = self.build_models(init_conditioner=False)
        self.first_stage_model = vae
        self.model = unet
        
        te1_sd, te2_sd, unet_sd, vae_sd = {}, {}, {}, {}
        for k in list(sd.keys()):
            if k.startswith("conditioner.embedders.0.transformer."):
                te1_sd[k.replace("conditioner.embedders.0.transformer.", "")] = sd.pop(k)
            elif k.startswith("conditioner.embedders.1.model."):
                te2_sd[k] = sd.pop(k)
            elif k.startswith("first_stage_model."):
                vae_sd[k.replace("first_stage_model.", "")] = sd.pop(k)
            elif k.startswith("model.diffusion_model."):
                unet_sd[k.replace("model.diffusion_model.", "diffusion_model.")] = sd.pop(k)
                
        converted_sd = convert_sdxl_text_encoder_2_checkpoint(te2_sd, max_length=77)
        self.text_encoder_1, self.text_encoder_2, self.tokenizer_1, self.tokenizer_2 = init_text_encoder()
        self.text_encoder_1.load_state_dict(te1_sd, strict=False)
        self.text_encoder_2.load_state_dict(converted_sd)
        vae.load_state_dict(vae_sd)
        unet.load_state_dict(unet_sd)
        self.noise_scheduler = DDPMScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            num_train_timesteps=1000,
            clip_sample=False,
        )
        self.init_lycoris()
        
    def init_lycoris(self):
        cfg = self.config 
        default_cfg = cfg.lycoris    
        lycoris_mapping = {}
        LycorisNetwork.apply_preset({"target_name": ".*diffusion_model.*"})
        self.lycoris_unet = create_lycoris(self.model, **cfg.get("lycoris_unet", default_cfg))
        self.lycoris_unet.to(self.target_device).apply_to()
        lycoris_mapping["unet"] = self.lycoris_unet
                
        LycorisNetwork.apply_preset({"target_name": ".*"})
        self.lycoris_te1 = create_lycoris(self.text_encoder_1, **cfg.get("lycoris_te1", default_cfg))
        self.lycoris_te1.to(self.target_device).apply_to()
        lycoris_mapping["te1"] = self.lycoris_te1

        LycorisNetwork.apply_preset({"target_name": ".*"})
        self.lycoris_te2 = create_lycoris(self.text_encoder_2, **cfg.get("lycoris_te2", default_cfg))
        self.lycoris_te2.to(self.target_device).apply_to()
        lycoris_mapping["te2"] = self.lycoris_te2
        
        self.lycoris_mapping = lycoris_mapping
        
    def encode_batch(self, batch):
        hidden1, hidden2, pooled = get_hidden_states_sdxl(
            batch["prompts"],
            self.max_token_length,
            self.tokenizer_1,
            self.tokenizer_2,
            self.text_encoder_1,
            self.text_encoder_2,
        )    
        emb = get_size_embeddings(
            batch["target_size_as_tuple"],
            batch["original_size_as_tuple"],
            batch["crop_coords_top_left"],
            self.fabric.device
        )
        cond = {
            "crossattn": torch.cat([hidden1, hidden2], dim=2),
            "vector": torch.cat([pooled, emb], dim=1),
        }
        return cond
        
    def save_checkpoint(self, model_path):
        cfg = self.config.trainer
        string_cfg = OmegaConf.to_yaml(self.config)
        state_dict = {}
        
        # build lycoris state_dict
        for key, module in self.lycoris_mapping.items():
            module_state_dict = module.state_dict()
            new_state_dict = {}
            for k, v in module_state_dict.items():
                k = k.replace("module.", "")
                if key == "unet":
                    k = k.replace("lycoris_diffusion_model_", "lycoris_unet_")
                elif key in ["te1", "te2"]:
                    k = k.replace("lycoris_", "")
                    k = f"lycoris_{key}_{k}"
                new_state_dict[k] = v
                
            state_dict.update(new_state_dict)
        
        if cfg.get("save_format") == "safetensors":
            model_path += ".safetensors"
            save_file(state_dict, model_path, metadata={"trainer_config": string_cfg})
        else:
            model_path += ".ckpt"
            torch.save(model_path, {"state_dict": self.model.state_dict()})
        rank_zero_print(f"Saved model to {model_path}")


from contextlib import nullcontext
from typing import Dict, List, Optional, Union

import math

import numpy as np
import open_clip
import torch
import torch.nn as nn
from einops import rearrange, repeat
from omegaconf import ListConfig
from torch.utils.checkpoint import checkpoint
from .model_util import (
    rank_zero_print,
    EmptyInitWrapper
)
from transformers import (
    CLIPTextModel,
    CLIPTextConfig,
    CLIPTokenizer,
    modeling_utils,
)
from .encoder_util import (
    count_params,
    default,
    disabled_train,
    expand_dims_like,
    instantiate_from_config,
)


def process_input_ids(input_ids, tokenizer, max_length):
    if max_length > tokenizer.model_max_length:
        input_ids = input_ids.squeeze(0)
        iids_list = []
        if tokenizer.pad_token_id == tokenizer.eos_token_id: # sdv1
            for i in range(1, max_length - tokenizer.model_max_length + 2, tokenizer.model_max_length - 2):  # (1, 152, 75)
                ids_chunk = (
                    input_ids[0].unsqueeze(0),
                    input_ids[i : i + tokenizer.model_max_length - 2],
                    input_ids[-1].unsqueeze(0),
                )
                ids_chunk = torch.cat(ids_chunk)
                iids_list.append(ids_chunk)
        else: # v2 or SDXL
            # 77以上の時は "<BOS> .... <EOS> <PAD> <PAD>..." でトータル227とかになっているので、"<BOS>...<EOS> <PAD> <PAD> ..."の三連に変換する
            for i in range(1, max_length - tokenizer.model_max_length + 2, tokenizer.model_max_length - 2):
                ids_chunk = (
                    input_ids[0].unsqueeze(0),  # BOS
                    input_ids[i : i + tokenizer.model_max_length - 2],
                    input_ids[-1].unsqueeze(0),
                )  # PAD or EOS
                ids_chunk = torch.cat(ids_chunk)
                
                # 末尾が <EOS> <PAD> または <PAD> <PAD> の場合は、何もしなくてよい
                # 末尾が x <PAD/EOS> の場合は末尾を <EOS> に変える（x <EOS> なら結果的に変化なし）
                if ids_chunk[-2] != tokenizer.eos_token_id and ids_chunk[-2] != tokenizer.pad_token_id:
                    ids_chunk[-1] = tokenizer.eos_token_id
                # 先頭が <BOS> <PAD> ... の場合は <BOS> <EOS> <PAD> ... に変える
                if ids_chunk[1] == tokenizer.pad_token_id:
                    ids_chunk[1] = tokenizer.eos_token_id

                iids_list.append(ids_chunk)

        input_ids = torch.stack(iids_list)  # 3,77
    return input_ids


def timestep_embedding(timesteps, dim, max_period=10000, repeat_only=False):
    """
    Create sinusoidal timestep embeddings.
    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    if not repeat_only:
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period)
            * torch.arange(start=0, end=half, dtype=torch.float32)
            / half
        ).to(device=timesteps.device)
        args = timesteps[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat(
                [embedding, torch.zeros_like(embedding[:, :1])], dim=-1
            )
    else:
        embedding = repeat(timesteps, "b -> b d", d=dim)
    return embedding


class Timestep(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, t):
        return timestep_embedding(t, self.dim)


class AbstractEmbModel(nn.Module):
    def __init__(self):
        super().__init__()
        self._is_trainable = None
        self._ucg_rate = None
        self._input_key = None

    @property
    def is_trainable(self) -> bool:
        return self._is_trainable

    @property
    def ucg_rate(self) -> Union[float, torch.Tensor]:
        return self._ucg_rate

    @property
    def input_key(self) -> str:
        return self._input_key

    @is_trainable.setter
    def is_trainable(self, value: bool):
        self._is_trainable = value

    @ucg_rate.setter
    def ucg_rate(self, value: Union[float, torch.Tensor]):
        self._ucg_rate = value

    @input_key.setter
    def input_key(self, value: str):
        self._input_key = value

    @is_trainable.deleter
    def is_trainable(self):
        del self._is_trainable

    @ucg_rate.deleter
    def ucg_rate(self):
        del self._ucg_rate

    @input_key.deleter
    def input_key(self):
        del self._input_key


class GeneralConditioner(nn.Module):
    OUTPUT_DIM2KEYS = {2: "vector", 3: "crossattn", 4: "concat", 5: "concat"}
    KEY2CATDIM = {"vector": 1, "crossattn": 2, "concat": 1}

    def __init__(self, emb_models: Union[List, ListConfig]):
        super().__init__()
        embedders = []
        for n, embconfig in enumerate(emb_models):
            embedder = instantiate_from_config(embconfig)
            embedder.is_trainable = embconfig.get("is_trainable", False)
            embedder.ucg_rate = embconfig.get("ucg_rate", 0.0)
            if not embedder.is_trainable:
                embedder.train = disabled_train
                if hasattr(embedder, "freeze"):
                    embedder.freeze()
                for param in embedder.parameters():
                    param.requires_grad = False
                embedder.eval()
            rank_zero_print(
                f"Initialized embedder #{n}: {embedder.__class__.__name__} "
                f"with {count_params(embedder, False)} params. Trainable: {embedder.is_trainable}"
            )

            if "input_key" in embconfig:
                embedder.input_key = embconfig["input_key"]
            elif "input_keys" in embconfig:
                embedder.input_keys = embconfig["input_keys"]
            else:
                raise KeyError(f"need either 'input_key' or 'input_keys' for embedder {embedder.__class__.__name__}")

            embedders.append(embedder)
        self.embedders = nn.ModuleList(embedders)

    def forward(self, batch: Dict, force_zero_embeddings=None):
        output = dict()
        if force_zero_embeddings is None:
            force_zero_embeddings = []
            
        for embedder in self.embedders:
            # print(embedder.__class__.__name__, embedder.input_key if hasattr(embedder, "input_key") else None)
            embedding_context = nullcontext if embedder.is_trainable else torch.no_grad
            with embedding_context():
                if hasattr(embedder, "input_key") and (embedder.input_key is not None):
                    emb_out = embedder(batch[embedder.input_key])
                elif hasattr(embedder, "input_keys"):
                    emb_out = embedder(*[batch[k] for k in embedder.input_keys])
                    
            assert isinstance(emb_out, (torch.Tensor, list, tuple)) \
                , f"encoder outputs must be tensors or a sequence, but got {type(emb_out)}"
            if not isinstance(emb_out, (list, tuple)):
                emb_out = [emb_out]
                
            for emb in emb_out:
                # print(emb.shape, emb.dim(), self.OUTPUT_DIM2KEYS[emb.dim()], self.KEY2CATDIM[self.OUTPUT_DIM2KEYS[emb.dim()]])
                out_key = self.OUTPUT_DIM2KEYS[emb.dim()]
                if embedder.ucg_rate > 0.0 and embedder.legacy_ucg_val is None:
                    emb = expand_dims_like(torch.bernoulli((1.0 - embedder.ucg_rate) * torch.ones(emb.shape[0], device=emb.device)),emb,) * emb
                    
                if hasattr(embedder, "input_key") and embedder.input_key in force_zero_embeddings:
                    emb = torch.zeros_like(emb)
                    
                if out_key in output:
                    output[out_key] = torch.cat((output[out_key], emb), self.KEY2CATDIM[out_key])
                else:
                    output[out_key] = emb
                    
        return output

    def get_unconditional_conditioning(
        self, batch_c, batch_uc=None, force_uc_zero_embeddings=None
    ):
        if force_uc_zero_embeddings is None:
            force_uc_zero_embeddings = []
        ucg_rates = list()
        for embedder in self.embedders:
            ucg_rates.append(embedder.ucg_rate)
            embedder.ucg_rate = 0.0
        c = self(batch_c)
        uc = self(batch_c if batch_uc is None else batch_uc, force_uc_zero_embeddings)

        for embedder, rate in zip(self.embedders, ucg_rates):
            embedder.ucg_rate = rate
        return c, uc


class IdentityEncoder(AbstractEmbModel):
    def encode(self, x):
        return x

    def forward(self, x):
        return x


class ClassEmbedder(AbstractEmbModel):
    def __init__(self, embed_dim, n_classes=1000, add_sequence_dim=False):
        super().__init__()
        self.embedding = nn.Embedding(n_classes, embed_dim)
        self.n_classes = n_classes
        self.add_sequence_dim = add_sequence_dim

    def forward(self, c):
        c = self.embedding(c)
        if self.add_sequence_dim:
            c = c[:, None, :]
        return c

    def get_unconditional_conditioning(self, bs, device="cuda"):
        uc_class = (
            self.n_classes - 1
        )  # 1000 classes --> 0 ... 999, one extra class for ucg (class 1000)
        uc = torch.ones((bs,), device=device) * uc_class
        uc = {self.key: uc.long()}
        return uc


class ClassEmbedderForMultiCond(ClassEmbedder):
    def forward(self, batch, key=None, disable_dropout=False):
        out = batch
        key = default(key, self.key)
        islist = isinstance(batch[key], list)
        if islist:
            batch[key] = batch[key][0]
        c_out = super().forward(batch, key, disable_dropout)
        out[key] = [c_out] if islist else c_out
        return out


class FrozenCLIPEmbedder(AbstractEmbModel):
    """Uses the CLIP transformer encoder for text (from huggingface)"""

    LAYERS = ["last", "pooled", "hidden"]

    def __init__(
        self,
        version="openai/clip-vit-large-patch14",
        device="cuda",
        max_length=227,
        layer="last",
        layer_idx=None,
        always_return_pooled=False,
    ):  # clip-vit-base-patch32
        super().__init__()
        assert layer in self.LAYERS
        self.tokenizer = CLIPTokenizer.from_pretrained(version)
        with modeling_utils.no_init_weights():
            config = CLIPTextConfig.from_pretrained(version)
            self.transformer = CLIPTextModel(config)
            # self.transformer = CLIPTextModel.from_pretrained(version)
                
        self.device = device
        self.max_length = max_length
        self.layer = layer
        self.layer_idx = layer_idx
        self.return_pooled = always_return_pooled
        if layer == "hidden":
            assert layer_idx is not None
            assert 0 <= abs(layer_idx) <= 12

    def freeze(self):
        self.transformer = self.transformer.eval()

        for param in self.parameters():
            param.requires_grad = False

    def forward(self, text):
        # not support pooled output at the moment
        assert not self.return_pooled and not self.layer == "pooled"
        
        batch_encoding = self.tokenizer(
            text,
            truncation=True,
            max_length=self.max_length,
            return_length=True,
            return_overflowing_tokens=False,
            padding="max_length",
            return_tensors="pt",
        ).input_ids
        
        # b,n,77
        input_ids = torch.stack([process_input_ids(batch, self.tokenizer, self.max_length) for batch in batch_encoding]).to(self.device)
        batch_size = input_ids.size()[0]
        
        # b,n,77 -> b*n, 77
        input_ids = input_ids.reshape((-1, self.tokenizer.model_max_length))
        outputs = self.transformer(input_ids=input_ids, output_hidden_states=self.layer == "hidden")
        
        if self.layer == "last":
            z = outputs.last_hidden_state
        else:
            z = outputs.hidden_states[self.layer_idx]
            
        # b*n, 77, 768 or 1280 -> b, n*77, 768 or 1280
        hidden_states = z.reshape((batch_size, -1, z.shape[-1]))
        states_list = [hidden_states[:, 0].unsqueeze(1)]  # <BOS>
        
        # <BOS> の後から <EOS> の前まで
        for i in range(1, self.max_length, self.tokenizer.model_max_length):
            states_list.append(hidden_states[:, i : i + self.tokenizer.model_max_length - 2])  
            
        states_list.append(hidden_states[:, -1].unsqueeze(1))  # <EOS>
        z = torch.cat(states_list, dim=1)       
        return z

    def encode(self, text):
        return self(text)


class OpenClipTextModel(open_clip.CLIP):
    output_dict: torch.jit.Final[bool]

    def __init__(
            self,
            embed_dim,
            vision_cfg,
            text_cfg,
            quick_gelu: bool = False,
            init_logit_scale: float = np.log(1 / 0.07),
            init_logit_bias: Optional[float] = None,
            cast_dtype: Optional[torch.dtype] = None,
            output_dict: bool = False,
    ):
        nn.Module.__init__(self)
        self.output_dict = output_dict
        text = open_clip.model._build_text_tower(embed_dim, text_cfg, quick_gelu, cast_dtype)
            
        self.transformer = text.transformer
        self.context_length = text.context_length
        self.vocab_size = text.vocab_size
        self.token_embedding = text.token_embedding
        self.positional_embedding = text.positional_embedding
        self.ln_final = text.ln_final
        self.text_projection = text.text_projection
        self.text_pool_type = text.pool_type
        self.register_buffer('attn_mask', text.attn_mask, persistent=False)

        self.logit_scale = nn.Parameter(torch.ones([]) * init_logit_scale)
        self.logit_bias = None
        if init_logit_bias is not None:
            self.logit_bias = nn.Parameter(torch.ones([]) * init_logit_bias)
            

class FrozenOpenCLIPEmbedder2(AbstractEmbModel):
    """
    Uses the OpenCLIP transformer encoder for text
    """

    LAYERS = ["pooled", "last", "penultimate"]

    def __init__(
        self,
        arch="ViT-H-14",
        version="laion2b_s32b_b79k",
        device="cuda",
        max_length=227,
        layer="last",
        always_return_pooled=False,
        legacy=True,
    ):
        super().__init__()
        assert layer in self.LAYERS
        with EmptyInitWrapper(device):
            model = OpenClipTextModel(**open_clip.get_model_config(arch))
            
        self.model = model
        self.device = device
        self.max_length = max_length
        self.return_pooled = always_return_pooled
        self.layer = layer
        if self.layer == "last":
            self.layer_idx = 0
        elif self.layer == "penultimate":
            self.layer_idx = 1
        else:
            raise NotImplementedError()
        self.legacy = legacy
        self.clip_tokenizer = CLIPTokenizer.from_pretrained(
            "laion/CLIP-ViT-bigG-14-laion2B-39B-b160k", 
            pad_token="!"
        ) # same as open clip tokenizer

    def freeze(self):
        self.model = self.model.eval()
        for param in self.parameters():
            param.requires_grad = False
            
    def process_hidden_state(self, z, batch_size):
        # b*n, 77, 768 or 1280 -> b, n*77, 768 or 1280
        hidden_states = z.reshape((batch_size, -1, z.shape[-1]))
        states_list = [hidden_states[:, 0].unsqueeze(1)]  # <BOS>
        
        # <BOS> の後から <EOS> の前まで
        for i in range(1, self.max_length, self.clip_tokenizer.model_max_length):
            states_list.append(hidden_states[:, i : i + self.clip_tokenizer.model_max_length - 2])  
            
        states_list.append(hidden_states[:, -1].unsqueeze(1))  # <EOS>
        z = torch.cat(states_list, dim=1)    
        return z

    def forward(self, text):
        tokens = open_clip.tokenize(text, context_length=self.max_length)
        input_ids = torch.stack([process_input_ids(batch, self.clip_tokenizer, self.max_length) for batch in tokens]).to(self.device)
        batch_size = input_ids.size()[0]
        
        input_ids = input_ids.reshape((-1, self.clip_tokenizer.model_max_length))
        z = self.encode_with_transformer(input_ids)
        if not self.return_pooled and self.legacy:
            return self.process_hidden_state(z, batch_size)
        
        hidden_state = z[self.layer]
        hidden_state = self.process_hidden_state(hidden_state, batch_size)
        if self.return_pooled:
            assert not self.legacy
            pooled = z["pooled"][::self.max_length // 75]
            return hidden_state, pooled
        
        return hidden_state

    def encode_with_transformer(self, text):
        x = self.model.token_embedding(text)  # [batch_size, n_ctx, d_model]
        x = x + self.model.positional_embedding
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.text_transformer_forward(x, attn_mask=self.model.attn_mask)
        if self.legacy:
            x = x[self.layer]
            x = self.model.ln_final(x)
            return x
        else:
            # x is a dict and will stay a dict
            o = x["last"]
            o = self.model.ln_final(o)
            pooled = self.pool(o, text)
            x["pooled"] = pooled
            return x

    def pool(self, x, text):
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = (
            x[torch.arange(x.shape[0]), text.argmax(dim=-1)]
            @ self.model.text_projection
        )
        return x

    def text_transformer_forward(self, x: torch.Tensor, attn_mask=None):
        outputs = {}
        for i, r in enumerate(self.model.transformer.resblocks):
            if i == len(self.model.transformer.resblocks) - 1:
                outputs["penultimate"] = x.permute(1, 0, 2)  # LND -> NLD
            if (
                self.model.transformer.grad_checkpointing
                and not torch.jit.is_scripting()
            ):
                x = checkpoint(r, x, attn_mask)
            else:
                x = r(x, attn_mask=attn_mask)
        outputs["last"] = x.permute(1, 0, 2)  # LND -> NLD
        return outputs

    def encode(self, text):
        return self(text)
    
class ConcatTimestepEmbedderND(AbstractEmbModel):
    """embeds each dimension independently and concatenates them"""

    def __init__(self, outdim):
        super().__init__()
        self.timestep = Timestep(outdim)
        self.outdim = outdim

    def forward(self, x):
        if x.ndim == 1:
            x = x[:, None]
        assert len(x.shape) == 2
        b, dims = x.shape[0], x.shape[1]
        x = rearrange(x, "b d -> (b d)")
        emb = self.timestep(x)
        emb = rearrange(emb, "(b d) d2 -> b (d d2)", b=b, d=dims, d2=self.outdim)
        return emb



import torch
import os
import safetensors.torch
import torch.utils._device
from lightning.pytorch.utilities import rank_zero_only
from common.logging import logger


@rank_zero_only
def rank_zero_print(*args, **kwargs):
    logger.info(*args, **kwargs)


@rank_zero_only
def rank_zero_warn(*args, **kwargs):
    logger.warning(*args, **kwargs)


@rank_zero_only
def rank_zero_debug(*args, **kwargs):
    logger.debug(*args, **kwargs)


class EmptyInitWrapper(torch.overrides.TorchFunctionMode):
    def __init__(self, device=None):
        self.device = device

    def __torch_function__(self, func, types, args=(), kwargs=None):
        kwargs = kwargs or {}
        if getattr(func, "__module__", None) == "torch.nn.init":
            if "tensor" in kwargs:
                return kwargs["tensor"]
            else:
                return args[0]
        if (
            self.device is not None
            and func in torch.utils._device._device_constructors()
            and kwargs.get("device") is None
        ):
            kwargs["device"] = self.device
        return func(*args, **kwargs)


def load_torch_file(ckpt, safe_load=False, device=None):
    if ckpt.lower().endswith(".safetensors"):
        sd = safetensors.torch.load_file(ckpt, device=device)
    else:
        if safe_load:
            if not "weights_only" in torch.load.__code__.co_varnames:
                print(
                    "Warning torch.load doesn't support weights_only on this pytorch version, loading unsafely."
                )
                safe_load = False
        if safe_load:
            pl_sd = torch.load(ckpt, map_location="cpu", weights_only=True)
        else:
            pl_sd = torch.load(ckpt, map_location="cpu")
        if "global_step" in pl_sd:
            rank_zero_print(f"Global Step: {pl_sd['global_step']}")
        if "state_dict" in pl_sd:
            sd = pl_sd["state_dict"]
        else:
            sd = pl_sd
    return sd


def save_torch_file(sd, ckpt, metadata=None):
    if metadata is not None:
        safetensors.torch.save_file(sd, ckpt, metadata=metadata)
    else:
        safetensors.torch.save_file(sd, ckpt)


def get_class(name: str):
    import importlib

    module_name, class_name = name.rsplit(".", 1)
    module = importlib.import_module(module_name, package=None)
    return getattr(module, class_name)


def setup_smddp(config):
    from lightning.pytorch.plugins.environments import LightningEnvironment

    # from lightning.fabric.strategies import DDPStrategy
    from common.fairscale import DDPShardedStrategy

    env = LightningEnvironment()
    env.world_size = lambda: int(os.environ["WORLD_SIZE"])
    env.global_rank = lambda: int(os.environ["RANK"])
    strategy = DDPShardedStrategy(
        cluster_environment=env,
        accelerator="gpu",
        static_graph=True,
    )

    world_size = int(os.environ["WORLD_SIZE"])
    num_gpus = int(os.environ["SM_NUM_GPUS"])
    num_nodes = int(world_size / num_gpus)
    init_params = {
        "devices": num_gpus,
        "num_nodes": num_nodes,
    }

    config.lightning.update(init_params)
    del config.lightning.accelerator

    config.trainer.checkpoint_dir = os.path.join(
        "/opt/ml/checkpoints", config.trainer.checkpoint_dir
    )
    config.sampling.save_dir = os.path.join(
        os.environ.get("SM_OUTPUT_DIR"), config.sampling.save_dir
    )
    return strategy, config


def get_latest_checkpoint(checkpoint_dir: str):
    if not os.path.isdir(checkpoint_dir):
        return None
    items = sorted(os.listdir(checkpoint_dir))
    if not items:
        return None
    return os.path.join(checkpoint_dir, items[-1])



import torch
import os
import lightning as pl
from omegaconf import OmegaConf
from datasets import load_dataset
from torchvision import transforms

import io
import copy
import torch.nn.functional as F
from PIL import Image
from common.utils import rank_zero_print, get_class
from modules.sdxl_model_diffusers import StableDiffusionModel
from lightning.pytorch.utilities.model_summary import ModelSummary



def setup(fabric: pl.Fabric, config: OmegaConf) -> tuple:
    model_path = config.trainer.model_path
    model = SupervisedFineTune(
        model_path=model_path, config=config, device=fabric.device
    )

    dataset, dataloader = setup_hf_dataloader(config)
    params_to_optim = [{"params": model.unet.parameters()}]
    # params_to_optim = [{'params': model.model.parameters()}]
    if config.advanced.get("train_text_encoder_1"):
        lr = config.advanced.get("text_encoder_1_lr", config.optimizer.params.lr)
        params_to_optim.append({"params": model.text_encoder_1.parameters(), "lr": lr})

    if config.advanced.get("train_text_encoder_2"):
        lr = config.advanced.get("text_encoder_2_lr", config.optimizer.params.lr)
        params_to_optim.append({"params": model.text_encoder_2.parameters(), "lr": lr})

    optim_param = config.optimizer.params
    optimizer = get_class(config.optimizer.name)(params_to_optim, **optim_param)
    scheduler = None
    if config.get("scheduler"):
        scheduler = get_class(config.scheduler.name)(
            optimizer, **config.scheduler.params
        )

    model.vae.to(torch.float32)
    if fabric.is_global_zero and os.name != "nt":
        print(f"\n{ModelSummary(model, max_depth=1)}\n")

    model, optimizer = fabric.setup(model, optimizer)
    dataloader = fabric.setup_dataloaders(dataloader)
    return model, dataset, dataloader, optimizer, scheduler


def setup_hf_dataloader(config):
    dataset = load_dataset(
        path=config.dataset.name,
        name=None,
        cache_dir=config.dataset.get("cache_dir", None),
    )
    reso = config.trainer.resolution
    interp = transforms.InterpolationMode.BILINEAR
    train_transforms = transforms.Compose(
        [
            transforms.Resize(reso, interpolation=interp),
            transforms.CenterCrop(reso),
            transforms.ToTensor(),
            transforms.Normalize([0.5], [0.5]),
        ]
    )

    def preprocess_train(examples):
        all_pixel_values = []
        for col_name in ["jpg_0", "jpg_1"]:
            images = [
                Image.open(io.BytesIO(im_bytes)).convert("RGB")
                for im_bytes in examples[col_name]
            ]
            original_sizes = [(image.height, image.width) for image in images]
            pixel_values = [train_transforms(image) for image in images]
            all_pixel_values.append(pixel_values)

        # Double on channel dim, jpg_y then jpg_w
        im_tup_iterator = zip(*all_pixel_values)
        combined_pixel_values = []
        for im_tup, label_0 in zip(im_tup_iterator, examples["label_0"]):
            if label_0 == 0:
                im_tup = im_tup[::-1]
            combined_im = torch.cat(im_tup, dim=0)  # no batch dim
            combined_pixel_values.append(combined_im)

        examples.update(
            {
                "original_size_as_tuple": original_sizes,
                "crop_coords_top_left": [(0, 0)] * len(original_sizes),
                "target_size_as_tuple": [(reso, reso)] * len(original_sizes),
                "prompts": examples["caption"],
                "pixels": combined_pixel_values,
            }
        )
        return examples

    def collate_fn(examples):
        pixel_values = torch.stack([e["pixels"] for e in examples])
        pixel_values = pixel_values.to(memory_format=torch.contiguous_format).float()
        orig_tup = [torch.tensor(e["original_size_as_tuple"]) for e in examples]
        crop_tup = [torch.tensor(e["crop_coords_top_left"]) for e in examples]
        target_tup = [torch.tensor(e["target_size_as_tuple"]) for e in examples]
        return_d = {
            "pixels": pixel_values,
            "original_size_as_tuple": torch.stack(orig_tup),
            "crop_coords_top_left": torch.stack(crop_tup),
            "target_size_as_tuple": torch.stack(target_tup),
            "prompts": [e["prompts"] for e in examples],
        }
        return return_d

    train_dataset = (
        dataset[config.dataset.dataset_split]
        .shuffle(seed=config.trainer.seed)
        .with_transform(preprocess_train)
    )
    dataloader = torch.utils.data.DataLoader(
        train_dataset,
        collate_fn=collate_fn,
        batch_size=config.trainer.batch_size,
        num_workers=4,
        drop_last=True,
    )
    return train_dataset, dataloader


class SupervisedFineTune(StableDiffusionModel):
    def init_model(self):
        super().init_model()
        # clone frozen unet as ref
        self.unet_ref = copy.deepcopy(self.unet)
        self.unet_ref.eval().requires_grad_(False)
        # since we're in dpo, unet_ref in 16bit is ok
        self.unet_ref.to(torch.float16)
        
    def forward(self, batch):
        advanced = self.config.get("advanced", {})
        self.vae.to(self.target_device)

        feed_pixel_values = torch.cat(batch["pixels"].chunk(2, dim=1))
        latents = self.encode_pixels(feed_pixel_values)
        if torch.any(torch.isnan(latents)):
            rank_zero_print("NaN found in latents, replacing with zeros")
            latents = torch.where(
                torch.isnan(latents), torch.zeros_like(latents), latents
            )

        self.text_encoder_1.to(self.target_device)
        self.text_encoder_2.to(self.target_device)

        model_dtype = next(self.unet.parameters()).dtype
        text_embedding, pooled = self.encode_prompt(batch)
        text_embedding = text_embedding.to(model_dtype).repeat(2, 1, 1)
        pooled = pooled.to(model_dtype).repeat(2, 1)

        bs = batch["original_size_as_tuple"]
        bc = batch["crop_coords_top_left"]
        bt = batch["target_size_as_tuple"]
        add_time_ids = torch.cat(
            [self.compute_time_ids(s, c, t) for s, c, t in zip(bs, bc, bt)]
        ).repeat(2, 1)

        # Sample noise that we'll add to the latents
        noise = torch.randn_like(latents, dtype=model_dtype).chunk(2)[0].repeat(2, 1, 1, 1)
        bsz = latents.shape[0] // 2

        # Sample a random timestep for each image
        timesteps = torch.randint(
            advanced.get("timestep_start", 0),
            advanced.get("timestep_end", 1000),
            (bsz,),
            dtype=torch.int64,
            device=latents.device,
        ).repeat(2)

        # Add noise to the latents according to the noise magnitude at each timestep
        # (this is the forward diffusion process)
        noisy_latents = self.noise_scheduler.add_noise(latents, noise, timesteps)

        # Predict the noise residual
        noise_pred = self.unet(
            sample=noisy_latents,
            timestep=timesteps,
            encoder_hidden_states=text_embedding,
            added_cond_kwargs={"text_embeds": pooled, "time_ids": add_time_ids},
        ).sample

        # Get the target for loss depending on the prediction type
        is_v = advanced.get("v_parameterization", False)
        target = (
            noise
            if not is_v
            else self.noise_scheduler.get_velocity(latents, noise, timesteps)
        )

        # Compute losses.
        model_losses = F.mse_loss(noise_pred.float(), target.float(), reduction="none")
        model_losses = model_losses.mean(dim=list(range(1, len(model_losses.shape))))
        model_losses_w, model_losses_l = model_losses.chunk(2)

        # For logging
        raw_model_loss = 0.5 * (model_losses_w.mean() + model_losses_l.mean())
        model_diff = model_losses_w - model_losses_l  # These are both LBS (as is t)

        with torch.no_grad():
            ref_preds = self.unet_ref(
                sample=noisy_latents,
                timestep=timesteps,
                encoder_hidden_states=text_embedding,
                added_cond_kwargs={"text_embeds": pooled, "time_ids": add_time_ids},
            ).sample

            ref_loss = F.mse_loss(ref_preds.float(), target.float(), reduction="none")
            ref_loss = ref_loss.mean(dim=list(range(1, len(ref_loss.shape))))

            ref_losses_w, ref_losses_l = ref_loss.chunk(2)
            ref_diff = ref_losses_w - ref_losses_l
            raw_ref_loss = ref_loss.mean()

        scale_term = -0.5 * self.config.beta_dpo
        inside_term = scale_term * (model_diff - ref_diff)
        loss = -1 * F.logsigmoid(inside_term).mean()

        implicit_acc = (inside_term > 0).sum().float() / inside_term.size(0)
        implicit_acc += 0.5 * (inside_term == 0).sum().float() / inside_term.size(0)

        self.fabric.log_dict(
            {
                "loss": loss.detach().item(),
                "raw_model_loss": raw_model_loss.detach().item(),
                "ref_loss": raw_ref_loss.detach().item(),
                "implicit_acc": implicit_acc.detach().item(),
            }
        )
        return loss



import functools
import hashlib
import json
import math
import h5py as h5
import numpy as np
import random
import torch

from tqdm.auto import tqdm
from dataclasses import dataclass
from pathlib import Path
from PIL import Image
from torch.utils.data import Dataset
from typing import Callable, Generator, Optional# type: ignore
from torchvision import transforms
from common.utils import rank_zero_debug, rank_zero_warn
from torchvision.transforms import Resize, InterpolationMode


image_suffix = set([".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".tif", ".webp"])

IMAGE_TRANSFORMS = transforms.Compose(
    [
        transforms.ToTensor(),
        transforms.Normalize([0.5], [0.5]),
    ]
)


def get_class(name: str):
    import importlib

    module_name, class_name = name.rsplit(".", 1)
    module = importlib.import_module(module_name, package=None)
    return getattr(module, class_name)


def is_img(path: Path):
    return path.suffix in image_suffix


def sha1sum(txt):
    return hashlib.sha1(txt.encode()).hexdigest()


@dataclass
class Entry:
    is_latent: bool
    pixel: torch.Tensor
    prompt: str
    original_size: tuple[int, int] # h, w
    cropped_size: Optional[tuple[int, int]] # h, w
    dhdw: Optional[tuple[int, int]] # dh, dw
    extras: dict = None
    # mask: torch.Tensor | None = None


def dirwalk(
    path: Path, cond: Optional[Callable] = None
) -> Generator[Path, None, None]:
    for p in path.iterdir():
        if p.is_dir():
            yield from dirwalk(p, cond)
        else:
            if isinstance(cond, Callable):
                if not cond(p):
                    continue
            yield p


class StoreBase(Dataset):
    """
    A dataset to prepare the instance and class images with the prompts for fine-tuning the model.
    It pre-processes the images and the tokenizes prompts.
    """

    def __init__(
        self,
        root_path,
        rank=0,
        dtype=torch.float16,
        use_central_crop=True,
        **kwargs,
    ):
        self.rank = rank
        self.root_path = Path(root_path)
        self.dtype = dtype
        self.use_central_crop = use_central_crop
        self.kwargs = kwargs

        self.length = 0
        self.rand_list: list = []
        self.to_ratio: None | np.ndarray = None
        self.raw_res: list[tuple[int, int]] = []
        self.curr_res: list[tuple[int, int]] = []
        
        assert self.root_path.exists()

    def get_raw_entry(self, index) -> tuple[bool, np.ndarray, str, (int, int)]:
        raise NotImplementedError

    def fix_aspect_randomness(self, rng: np.random.Generator):
        raise NotImplementedError
    
    @staticmethod
    @functools.cache
    def fit_dimensions(target_ratio, min_h, min_w):
        min_area = min_h * min_w
        h = max(min_h, math.ceil(math.sqrt(min_area * target_ratio)))
        w = max(min_w, math.ceil(h / target_ratio))
        
        if w < min_w:
            w = min_w
            h = max(min_h, math.ceil(w * target_ratio))
        
        while h * w < min_area:
            increment = 8
            if target_ratio >= 1:
                h += increment
            else:
                w += increment

            w = max(min_w, math.ceil(h / target_ratio))
            h = max(min_h, math.ceil(w * target_ratio))
        return int(h), int(w)

    def crop(self, entry: Entry, i: int) -> Entry:
        assert self.to_ratio is not None, "to_ratio is not initialized"
        H, W = entry.pixel.shape[-2:]
        base_ratio = H / W
        target_ratio = self.to_ratio[i]
        h, w = self.ratio_to_bucket[target_ratio]
        if not entry.is_latent:
            resize_h, resize_w = self.fit_dimensions(base_ratio, h, w)
            interp = InterpolationMode.BILINEAR if resize_h < H else InterpolationMode.BICUBIC
            entry.pixel = Resize((resize_h, resize_w), interpolation=interp, antialias=None)(entry.pixel)
        else:
            h, w = h // 8 , w // 8 
        
        H, W = entry.pixel.shape[-2:]
        if self.use_central_crop:
            dh, dw = (H - h) // 2, (W - w) // 2
        else:
            assert H >= h and W >= w, f"{H}<{h} or {W}<{w}"
            dh, dw = random.randint(0, H - h), random.randint(0, W - w)
            
        entry.pixel = entry.pixel[:, dh : dh + h, dw : dw + w]
        return entry, dh, dw

    @torch.no_grad()
    def get_batch(self, indices: list[int]) -> Entry:
        entries = [self._get_entry(i) for i in indices]
        crop_pos = []
        pixels = []
        prompts = []
        original_sizes = []
        cropped_sizes = []
        extras = []

        for e, i in zip(entries, indices):
            e, dh, dw = self.crop(e, i)
            pixels.append(e.pixel)
            original_size = torch.asarray(e.original_size)
            original_sizes.append(original_size)

            cropped_size = e.pixel.shape[-2:]
            cropped_size = (cropped_size[0] * 8, cropped_size[1] * 8) if e.is_latent else cropped_size
            cropped_size = torch.asarray(cropped_size)
            cropped_sizes.append(cropped_size)
            
            cropped_pos = (dh, dw)
            cropped_pos = (cropped_pos[0] * 8, cropped_pos[1] * 8) if e.is_latent else cropped_pos
            cropped_pos = (cropped_pos[0] + e.dhdw[0], cropped_pos[1] + e.dhdw[1])
            cropped_pos = torch.asarray(cropped_pos)
            crop_pos.append(cropped_pos)
            prompts.append(e.prompt)
            extras.append(self.process_batch_extras(e.extras))
            
        is_latent = entries[0].is_latent
        shape = entries[0].pixel.shape

        for e in entries[1:]:
            assert e.is_latent == is_latent
            assert e.pixel.shape == shape, f"{e.pixel.shape} != {shape} for the same batch"

        pixel = torch.stack(pixels, dim=0).contiguous()
        cropped_sizes = torch.stack(cropped_sizes)
        original_sizes = torch.stack(original_sizes)
        crop_pos = torch.stack(crop_pos)

        return {
            "prompts": prompts,
            "pixels": pixel,
            "is_latent": is_latent,
            "target_size_as_tuple": cropped_sizes,
            "original_size_as_tuple": original_sizes,
            "crop_coords_top_left": crop_pos,
            "extras": extras,
        }

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        raise NotImplementedError
    
    def get_batch_extras(self, path):
        return None
    
    def process_batch_extras(self, inputs):
        return inputs

    def _get_entry(self, index) -> Entry:
        is_latent, pixel, prompt, original_size, dhdw, extras = self.get_raw_entry(index)   
        pixel = pixel.to(dtype=self.dtype)
        shape = pixel.shape
        if shape[-1] == 3 and shape[-1] < shape[0] and shape[-1] < shape[1]:  
            pixel = pixel.permute(2, 0, 1) # HWC -> CHW
            
        return Entry(is_latent, pixel, prompt, original_size, None, dhdw, extras)
    
    def repeat_entries(self, k, res, index=None):
        repeat_strategy = self.kwargs.get("repeat_strategy", None)
        if repeat_strategy is not None:
            assert index is not None
            index_new = index.copy()
            for i, ent in enumerate(index):
                for strategy, mult in repeat_strategy:
                    if strategy in str(ent):
                        k.extend([k[i]] * (mult-1))
                        res.extend([res[i]] * (mult-1))
                        index_new.extend([index_new[i]] * (mult-1))
                        break
        else:
            index_new = index
        return k, res, index_new


class LatentStore(StoreBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        prompt_mapping = next(dirwalk(self.root_path, lambda p: p.suffix == ".json"))
        prompt_mapping = json.loads(Path(prompt_mapping).read_text())
        
        self.h5_paths = list(dirwalk(self.root_path, lambda p: p.suffix == ".h5" and "prompt_cache" not in p.stem))
        self.h5_keymap = {}
        self.h5_filehandles = {}
        self.paths = []
        total_latents = len(self.h5_paths)
        for idx, h5_path in enumerate(self.h5_paths):
            fs = h5.File(h5_path, "r", libver="latest")
            for k in tqdm(fs.keys(), desc=f"Loading latents {idx+1}/{total_latents}", disable=self.rank != 0, leave=False, ascii=True):
                hashkey = k[:-8] # ".latents"
                assert hashkey in prompt_mapping, f"Key {k} not found in prompt_mapping"
                it = prompt_mapping[hashkey]
                if not it["train_use"]: continue
                prompt, it_path = it["train_caption"], it["file_path"]
                height, width = it["train_height"], it["train_width"]
                self.paths.append(it_path)
                self.raw_res.append((height, width))
                self.h5_keymap[k] = (h5_path, prompt, (height, width))
        
        self.keys = list(self.h5_keymap.keys())
        self.length = len(self.keys)
        rank_zero_debug(f"Loaded {self.length} latent codes from {self.root_path}")
        
        self.keys, self.raw_res, self.paths = self.repeat_entries(self.keys, self.raw_res, index=self.paths)
        new_length = len(self.paths)
        if new_length != self.length:
            self.length = new_length
            rank_zero_debug(f"Using {self.length} entries after applied repeat strategy")
        
    def setup_filehandles(self):
        self.h5_filehandles = {}
        for h5_path in self.h5_paths:
            self.h5_filehandles[h5_path] = h5.File(h5_path, "r", libver="latest")

    def get_raw_entry(self, index) -> tuple[bool, torch.tensor, str, (int, int)]:
        if len(self.h5_filehandles) == 0:
            self.setup_filehandles()
        latent_key = self.keys[index]
        h5_path, prompt, original_size = self.h5_keymap[latent_key]
        latent = torch.asarray(self.h5_filehandles[h5_path][latent_key][:]).float()
        scaled = self.h5_filehandles[h5_path][latent_key].attrs.get("scale", True)
        dhdw = self.h5_filehandles[h5_path][latent_key].attrs.get("dhdw", (0, 0))
        latent = latent * 0.13025 if not scaled else latent
        
        extras = self.get_batch_extras(self.paths[index])
        return True, latent, prompt, original_size, dhdw, extras


class DirectoryImageStore(StoreBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        label_ext = self.kwargs.get("label_ext", ".txt")
        self.paths = list(dirwalk(self.root_path, is_img))
        self.length = len(self.paths)
        rank_zero_debug(f"Found {self.length} images in {self.root_path}")

        remove_paths = []
        for p in tqdm(
            self.paths, desc="Loading image sizes", leave=False, ascii=True,
        ):
            try:
                w, h = Image.open(p).size
                self.raw_res.append((h, w))
            except Exception as e:
                print(f"\033[33mSkipped: error processing {p}: {e}\033[0m")
                remove_paths.append(p)

        remove_paths = set(remove_paths)
        self.paths = [p for p in self.paths if p not in remove_paths]  
        self.length = len(self.raw_res)
        
        self.length = len(self.paths)
        self.prompts: list[str] = []
        for path in tqdm(
            self.paths,
            desc="Loading prompts",
            disable=self.rank != 0,
            leave=False,
            ascii=True,
        ):
            p = path.with_suffix(label_ext)
            try:
                with open(p, "r") as f:
                    self.prompts.append(f.read())
            except Exception as e:
                rank_zero_warn(f"Skipped: error processing {p}: {e}")
                self.prompts.append("")
            
        self.base_len = self.kwargs["base_len"]
        rank_zero_debug(f"Loaded {len(self.prompts)} prompts, {self.length} image sizes")
        
        self.prompts, self.raw_res, self.paths = self.repeat_entries(self.prompts, self.raw_res, index=self.paths)
        new_length = len(self.paths)
        if new_length != self.length:
            self.length = new_length
            rank_zero_debug(f"Using {self.length} entries after applied repeat strategy")

    def get_raw_entry(self, index) -> tuple[bool, torch.tensor, str, (int, int)]:
        p = self.paths[index]
        prompt = self.prompts[index]
        _img = Image.open(p)
        if _img.mode == "RGB":
            img = np.array(_img)
        elif _img.mode == "RGBA":
            img = np.array(_img)
            rgb, alpha = img[:, :, :3], img[:, :, 3:]
            fp_alpha = alpha / 255
            rgb[:] = rgb * fp_alpha + (255 - alpha)
            img = rgb
        else:
            img = np.array(_img.convert("RGB"))
            
        img = IMAGE_TRANSFORMS(img)
        h, w = img.shape[-2:]
        dhdw = (0, 0)
        extras = self.get_batch_extras(p)
        return False, img, prompt, (h, w), dhdw, extras


import torch
import torch.utils.checkpoint
import lightning as pl

from pathlib import Path
from common.utils import rank_zero_print
from diffusers import StableDiffusionPipeline
from diffusers import DDIMScheduler, DDPMScheduler
from lightning.pytorch.utilities import rank_zero_only

# define the LightningModule
class StableDiffusionModel(pl.LightningModule):
    def __init__(self, model_path, config, device):
        super().__init__()
        self.config = config
        self.model_path = model_path
        self.target_device = device
        self.init_model()

    def init_model(self):
        trainer_cfg = self.config.trainer
        config = self.config
        advanced = config.get("advanced", {})
        
        rank_zero_print(f"Loading model from {self.model_path}")
        p = StableDiffusionPipeline
        if Path(self.model_path).is_dir():
            self.pipeline = pipeline = p.from_pretrained(self.model_path)
        else:
            self.pipeline = pipeline = p.from_single_file(self.model_path)
            
        self.vae, self.unet = pipeline.vae, pipeline.unet
        self.text_encoder, self.tokenizer = (
            pipeline.text_encoder,
            pipeline.tokenizer,
        )
        self.max_prompt_length = 225 + 2
        self.vae.requires_grad_(False)
        self.text_encoder.requires_grad_(False)
        self.scale_factor = 0.13025

        self.unet.enable_gradient_checkpointing()
        if trainer_cfg.use_xformers:
            self.unet.enable_xformers_memory_efficient_attention()

        if advanced.get("train_text_encoder"):
            self.text_encoder.requires_grad_(True)
            self.text_encoder.gradient_checkpointing_enable()

        self.noise_scheduler = DDPMScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            num_train_timesteps=1000,
            clip_sample=False,
        )

    def encode_prompts(self, prompts):            
        input_ids = self.tokenizer(prompts, padding="do_not_pad", truncation=True, max_length=225).input_ids 
        input_ids = self.tokenizer.pad({"input_ids": input_ids}, padding=True, return_tensors="pt").input_ids
            
        z = []
        if input_ids.shape[1] > 77:
            # todo: Handle end-of-sentence truncation
            while max(map(len, input_ids)) != 0:
                rem_tokens = [x[75:] for x in input_ids]
                tokens = []
                for j in range(len(input_ids)):
                    tokens.append(input_ids[j][:75] if len(input_ids[j]) > 0 else [self.tokenizer.eos_token_id] * 75)

                rebuild = [[self.tokenizer.bos_token_id] + list(x[:75]) + [self.tokenizer.eos_token_id] for x in tokens]
                if hasattr(torch, "asarray"):
                    z.append(torch.asarray(rebuild))
                else:
                    z.append(torch.IntTensor(rebuild))
                input_ids = rem_tokens
        else:
            z.append(input_ids)

        # Get the text embedding for conditioning
        encoder_hidden_states = None
        for tokens in z:
            state = self.text_encoder(tokens.to(self.device), output_hidden_states=True)
            state = self.text_encoder.text_model.final_layer_norm(state['hidden_states'][-self.config.trainer.clip_skip])
            encoder_hidden_states = state if encoder_hidden_states is None else torch.cat((encoder_hidden_states, state), axis=-2)

        return encoder_hidden_states

    @torch.inference_mode()
    @rank_zero_only
    def sample(
        self,
        prompt,
        negative_prompt="lowres, low quality, text, error, extra digit, cropped",
        generator=None,
        size=(512, 512),
        steps=20,
        guidance_scale=6.5,
    ):
        self.vae.to(self.target_device)
        scheduler = DDIMScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            num_train_timesteps=1000,
        )
        pipeline = StableDiffusionPipeline(
            unet=self.unet,
            vae=self.vae,
            text_encoder_1=self.text_encoder_1,
            text_encoder_2=self.text_encoder_2,
            noise_scheduler=scheduler,
        )
        image = pipeline(
            prompt=prompt,
            negative_prompt=negative_prompt,
            width=size[0],
            height=size[1],
            steps=steps,
            generator=generator,
            guidance_scale=guidance_scale,
            return_dict=False,
        )[0]
        return image

    def save_checkpoint(self, model_path):
        self.pipeline.save_pretrained(model_path)
        rank_zero_print(f"Saved model to {model_path}")

    def forward(self, batch):
        raise NotImplementedError



import numpy as np
import random
import torch

from pathlib import Path
from torch.utils.data import Dataset, get_worker_info
from common.storage import DirectoryImageStore, LatentStore

image_suffix = set([".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".tif", ".webp"])


def is_latent_folder(path: Path):
    # iterate over all files in the folder and find if any of them is a latent
    for p in path.iterdir():
        if p.is_dir():
            continue
        if p.suffix == ".h5":
            return True


class AspectRatioDataset(Dataset):
    def __init__(
        self,
        batch_size: int,
        img_path: Path | str | list,
        ucg: int = 0,
        rank: int = 0,
        dtype=torch.float16,
        target_area: int = 1024 * 1024,
        min_size: int = 512,
        max_size: int = 2048,
        divisible: int = 64,
        seed: int = 42,
        **kwargs,
    ):
        self.rng = np.random.default_rng(seed)
        self.batch_size = batch_size

        root_path = Path(img_path)
        assert root_path.exists()

        store_class = DirectoryImageStore
        if is_latent_folder(root_path):
            store_class = LatentStore

        self.store = store_class(
            root_path,
            rank=rank,
            ucg=ucg,
            dtype=dtype,
            **kwargs,
        )
        self.target_area = target_area
        self.max_size, self.min_size, self.divisible = max_size, min_size, divisible
        self.first_time = True
        self.generate_buckets()
        self.init_batches()

    def init_batches(self):
        self.assign_buckets()
        self.assign_batches()

    def __len__(self):
        return len(self.batch_idxs)

    def generate_buckets(self):
        assert (
            self.target_area % 4096 == 0
        ), "target area (h * w) must be divisible by 64"
        width = np.arange(self.min_size, self.max_size + 1, self.divisible)
        height = np.minimum(self.max_size, ((self.target_area // width) // self.divisible) * self.divisible,)
        valid_mask = height >= self.min_size

        resos = set(zip(width[valid_mask], height[valid_mask]))
        resos.update(zip(height[valid_mask], width[valid_mask]))
        resos.add(((int(np.sqrt(self.target_area)) // self.divisible) * self.divisible,) * 2)
        self.buckets_sizes = np.array(sorted(resos))
        self.bucket_ratios = self.buckets_sizes[:, 0] / self.buckets_sizes[:, 1]
        self.store.ratio_to_bucket = {ratio: hw for ratio, hw in zip(self.bucket_ratios, self.buckets_sizes)}

    def assign_buckets(self):
        img_res = np.array(self.store.raw_res)
        img_ratios = img_res[:, 0] / img_res[:, 1]
        self.bucket_content = [[] for _ in range(len(self.buckets_sizes))]
        self.store.to_ratio = {}

        # Assign images to buckets
        for idx, img_ratio in enumerate(img_ratios):
            diff = np.abs(self.bucket_ratios - img_ratio)
            bucket_idx = np.argmin(diff)
            self.bucket_content[bucket_idx].append(idx)
            self.store.to_ratio[idx] = self.bucket_ratios[bucket_idx]

    def assign_batches(self):
        self.batch_idxs = []
        for bucket in self.bucket_content:
            if not bucket:
                continue
            reminder = len(bucket) % self.batch_size
            bucket = np.array(bucket)
            self.rng.shuffle(bucket)
            if not reminder:
                self.batch_idxs.extend(bucket.reshape(-1, self.batch_size))
            else:
                self.batch_idxs.extend(bucket[:-reminder].reshape(-1, self.batch_size))
                self.batch_idxs.append(bucket[-reminder:])

        np.random.shuffle(self.batch_idxs)

    def __getitem__(self, idx):
        img_idxs = self.batch_idxs[idx]
        return self.store.get_batch(img_idxs)


def worker_init_fn(worker_id):
    worker_info = get_worker_info()
    dataset: AspectRatioDataset = worker_info.dataset  # type: ignore
    random.seed(worker_info.seed)  # type: ignore
    dataset.init_batches()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# GLIDE: https://github.com/openai/glide-text2im
# MAE: https://github.com/facebookresearch/mae/blob/main/models_mae.py
# --------------------------------------------------------
import math, os
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import functools

from einops import rearrange
from tqdm.auto import tqdm

XFORMERS_AVAILABLE = False
try:
    import xformers.ops
    XFORMERS_AVAILABLE = True
except ImportError:
    pass

def modulate(x, shift, scale):
    return x * (1 + scale) + shift

class Mlp(nn.Module):
    """ MLP as used in Vision Transformer, MLP-Mixer and related networks
    """
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Linear(in_features, hidden_features, bias=True)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features, bias=True)

    def forward(self, x):
        return self.fc2(self.act(self.fc1(x)))


class PatchEmbed(nn.Module):
    """2D Image to Patch Embedding"""

    def __init__(
        self,
        input_size,
        patch_size=16,
        in_channels=3,
        embed_dim=768,
        flatten=True,
        bias=True,
        interp_scale=2,
    ):
        super().__init__()
        height = width = input_size
        num_patches = (height // patch_size) * (width // patch_size)
        self.flatten = flatten

        self.proj = nn.Conv2d(
            in_channels, embed_dim, kernel_size=(patch_size, patch_size), stride=patch_size, bias=bias
        )

        self.patch_size = patch_size
        self.height, self.width = height // patch_size, width // patch_size
        self.base_size = height // patch_size
        self.interpolation_scale = interp_scale
        pos_embed = get_2d_sincos_pos_embed(
            embed_dim, int(num_patches**0.5), base_size=self.base_size, interpolation_scale=self.interpolation_scale
        )
        self.register_buffer("pos_embed", torch.from_numpy(pos_embed).float().unsqueeze(0), persistent=False)
        
    @functools.cache
    def get_pos_embed(self, latent_height, latent_width):
        height, width = latent_height // self.patch_size, latent_width // self.patch_size
        if self.height != height or self.width != width:
            pos_embed = get_2d_sincos_pos_embed(
                embed_dim=self.pos_embed.shape[-1],
                grid_size=(height, width),
                base_size=self.base_size,
                interpolation_scale=self.interpolation_scale,
            )
            pos_embed = torch.from_numpy(pos_embed)
            pos_embed = pos_embed.float().unsqueeze(0)
        else:
            pos_embed = self.pos_embed
        return pos_embed

    def forward(self, latent, pos_embed):
        # height, width = latent.shape[-2] // self.patch_size, latent.shape[-1] // self.patch_size

        latent = self.proj(latent)
        latent = latent.flatten(2).transpose(1, 2)  # BCHW -> BNC

        # Interpolate positional embeddings if needed.
        # if precomputed_pos_embed is not None:
        #     pos_embed = precomputed_pos_embed
        # elif self.height != height or self.width != width:
        #     pos_embed = get_2d_sincos_pos_embed(
        #         embed_dim=self.pos_embed.shape[-1],
        #         grid_size=(height, width),
        #         base_size=self.base_size,
        #         interpolation_scale=self.interpolation_scale,
        #     )
        #     pos_embed = torch.from_numpy(pos_embed)
        #     pos_embed = pos_embed.float().unsqueeze(0).to(latent.device)
        # else:
        #     pos_embed = self.pos_embed
        return (latent + pos_embed).to(latent.dtype)
    

class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False
    ):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
    
    def forward(self, x):
        B, N, C = x.shape
        
        if XFORMERS_AVAILABLE:
            qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 1, 3, 4)
            q, k, v = qkv.unbind(0)
            x = xformers.ops.memory_efficient_attention(q, k, v)
            x = x.reshape(B, N, C)
        else:    
            qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
            q, k, v = qkv.unbind(0)
            x = F.scaled_dot_product_attention(q, k, v)
            x = x.transpose(1, 2).reshape(B, N, C)
        
        x = self.proj(x)
        return x


class MultiHeadCrossAttention(nn.Module):
    def __init__(
        self, 
        d_model, 
        num_heads
    ):
        super(MultiHeadCrossAttention, self).__init__()
        assert d_model % num_heads == 0, "d_model must be divisible by num_heads"

        self.d_model = d_model
        self.num_heads = num_heads
        self.head_dim = d_model // num_heads

        self.q_linear = nn.Linear(d_model, d_model)
        self.kv_linear = nn.Linear(d_model, d_model*2)
        self.proj = nn.Linear(d_model, d_model)

    def forward(self, x, cond, attention_mask=None):
        # query/value: img tokens; key: condition; mask: if padding tokens
        B, N, C = x.shape
        
        if attention_mask is not None:
            attention_mask = attention_mask.repeat_interleave(self.num_heads, dim=0)
            attention_mask = attention_mask.view(B, self.num_heads, -1, attention_mask.shape[-1]).to(x.dtype)
        
        q = self.q_linear(x).view(B, -1, self.num_heads, self.head_dim)
        kv = self.kv_linear(cond).view(B, -1, 2, self.num_heads, self.head_dim)
        k, v = kv.unbind(2)
        
        if XFORMERS_AVAILABLE:
            attention_mask = attention_mask.expand(-1, -1, N, -1).to(q.dtype)
            x = xformers.ops.memory_efficient_attention(q, k, v, attn_bias=attention_mask)
            x = x.reshape(B, N, C)
            x = self.proj(x)
        else:
            q, k, v = q.transpose(1, 2), k.transpose(1, 2), v.transpose(1, 2)
            x = F.scaled_dot_product_attention(q, k, v, attn_mask=attention_mask)
            x = x.transpose(1, 2).reshape(B, N, C)
            x = self.proj(x)
        
        return x


class FinalLayer(nn.Module):
    def __init__(self, hidden_size, patch_size, out_channels):
        super().__init__()
        self.norm_final = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.linear = nn.Linear(hidden_size, patch_size * patch_size * out_channels, bias=True)
        self.scale_shift_table = nn.Parameter(torch.randn(2, hidden_size) / hidden_size ** 0.5)
        self.out_channels = out_channels

    def forward(self, x, t):
        shift, scale = (self.scale_shift_table[None] + t[:, None]).chunk(2, dim=1)
        # x = modulate(self.norm_final(x), shift, scale)
        x = self.norm_final(x) * (1 + scale) + shift
        x = self.linear(x)
        return x


#################################################################################
#               Embedding Layers for Timesteps and Class Labels                 #
#################################################################################

class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """
    def __init__(self, hidden_size, frequency_embedding_size=256):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000.0):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """
        # https://github.com/openai/glide-text2im/blob/main/glide_text2im/nn.py
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=t.device)
        
        args = t[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t, dtype):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq.to(dtype))
        return t_emb


class CaptionEmbedder(nn.Module):
    """
    Embeds class labels into vector representations. Also handles label dropout for classifier-free guidance.
    """

    def __init__(self, in_channels, hidden_size, uncond_prob, act_layer=nn.GELU(approximate='tanh'), token_num=120):
        super().__init__()
        self.y_proj = Mlp(in_features=in_channels, hidden_features=hidden_size, out_features=hidden_size, act_layer=act_layer)
        self.register_buffer("y_embedding", nn.Parameter(torch.randn(token_num, in_channels) / in_channels ** 0.5))
        self.uncond_prob = uncond_prob

    def forward(self, caption):
        caption = self.y_proj(caption)
        return caption
    
    
class SizeEmbedder(TimestepEmbedder):
    """
    Embeds scalar timesteps into vector representations.
    """

    def __init__(self, hidden_size, frequency_embedding_size=256):
        super().__init__(hidden_size=hidden_size, frequency_embedding_size=frequency_embedding_size)
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size
        self.outdim = hidden_size

    def forward(self, s, bs, dtype):
        if s.ndim == 1:
            s = s[:, None]
        assert s.ndim == 2
        if s.shape[0] != bs:
            s = s.repeat(bs//s.shape[0], 1)
            assert s.shape[0] == bs
        b, dims = s.shape[0], s.shape[1]
        s = rearrange(s, "b d -> (b d)")
        s_freq = self.timestep_embedding(s, self.frequency_embedding_size)
        s_emb = self.mlp(s_freq.to(dtype))
        s_emb = rearrange(s_emb, "(b d) d2 -> b (d d2)", b=b, d=dims, d2=self.outdim)
        return s_emb


class DiTBlock(nn.Module):
    """
    A DiT block with adaptive layer norm (adaLN-single) conditioning.
    """

    def __init__(self, hidden_size, num_heads, mlp_ratio=4.0, **block_kwargs):
        super().__init__()
        self.norm1 = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.attn = Attention(hidden_size, num_heads=num_heads, qkv_bias=True)
        self.mlp = Mlp(in_features=hidden_size, hidden_features=int(hidden_size * mlp_ratio), act_layer=lambda: nn.GELU(approximate="tanh"))
        self.norm2 = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.cross_attn = MultiHeadCrossAttention(hidden_size, num_heads, **block_kwargs)
        # self.xattn_norm = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.scale_shift_table = nn.Parameter(torch.randn(6, hidden_size) / hidden_size ** 0.5)

    def forward(self, x, t, y, attention_mask):
        B, N, C = x.shape
        shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp = (self.scale_shift_table[None] + t.reshape(B, 6, -1)).chunk(6, dim=1)
            
        x = x + gate_msa * self.attn(modulate(self.norm1(x), shift_msa, scale_msa)).reshape(B, N, C)
        x = x + self.cross_attn(x, y, attention_mask)
        x = x + gate_mlp * self.mlp(modulate(self.norm2(x), shift_mlp, scale_mlp))
        return x
    
def ckpt_wrapper(module, *inputs):
    def ckpt_forward(*inputs):
        outputs = module(*inputs)
        return outputs
    return torch.utils.checkpoint.checkpoint(ckpt_forward, *inputs, use_reentrant=False)


#############################################################################
#                                 Core DiT Model                                #
#################################################################################
class DiTModel(nn.Module):
    """
    Diffusion model with a Transformer backbone.
    """

    def __init__(
        self,
        input_size=32,
        patch_size=2,
        in_channels=4,
        hidden_size=1152,
        depth=28,
        num_heads=16,
        mlp_ratio=4.0,
        class_dropout_prob=0.1,
        caption_channels=4096,
        learn_sigma=True,
        interpolation_scale=2.,
    ):
        super().__init__()
        self.learn_sigma = learn_sigma
        self.in_channels = in_channels
        self.out_channels = in_channels * 2 if learn_sigma else in_channels
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.base_size = input_size // self.patch_size

        self.x_embedder = PatchEmbed(input_size, patch_size, in_channels, hidden_size, bias=True, interp_scale=interpolation_scale)
        self.t_embedder = TimestepEmbedder(hidden_size)
        
        self.csize_embedder = SizeEmbedder(hidden_size//3)  # c_size embed
        self.ar_embedder = SizeEmbedder(hidden_size//3)     # aspect ratio embed
        approx_gelu = lambda: nn.GELU(approximate="tanh")
        
        # adaln single fn (= adaln_single.linear)
        self.t_block = nn.Sequential(
            nn.SiLU(),
            nn.Linear(hidden_size, 6 * hidden_size, bias=True)
        )
        self.y_embedder = CaptionEmbedder(
            in_channels=caption_channels, 
            hidden_size=hidden_size, 
            uncond_prob=class_dropout_prob, 
            act_layer=approx_gelu,
            token_num=120,
        )
        self.blocks = nn.ModuleList([
            DiTBlock(hidden_size, num_heads, mlp_ratio=mlp_ratio) for _ in range(depth)
        ])
        self.final_layer = FinalLayer(hidden_size, patch_size, self.out_channels)

    def initialize_weights(self):
        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)
        self.apply(_basic_init)

        # Initialize patch_embed like nn.Linear (instead of nn.Conv2d):
        w = self.x_embedder.proj.weight.data
        nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

        # Initialize timestep embedding MLP:
        nn.init.normal_(self.t_embedder.mlp[0].weight, std=0.02)
        nn.init.normal_(self.t_embedder.mlp[2].weight, std=0.02)
        nn.init.normal_(self.t_block[1].weight, std=0.02)

        # Initialize caption embedding MLP:
        nn.init.normal_(self.y_embedder.y_proj.fc1.weight, std=0.02)
        nn.init.normal_(self.y_embedder.y_proj.fc2.weight, std=0.02)

        # Zero-out adaLN modulation layers in DiT blocks:
        for block in self.blocks:
            nn.init.constant_(block.cross_attn.proj.weight, 0)
            nn.init.constant_(block.cross_attn.proj.bias, 0)

        # Zero-out output layers:
        nn.init.constant_(self.final_layer.linear.weight, 0)
        nn.init.constant_(self.final_layer.linear.bias, 0)

    def ckpt_wrapper(self, module):
        def ckpt_forward(*inputs):
            outputs = module(*inputs)
            return outputs
        return ckpt_forward
    
    def unpatchify(self, x):
        """
        x: (N, T, patch_size**2 * C)
        imgs: (N, H, W, C)
        """
        c = self.out_channels
        p = self.x_embedder.patch_size
        # h = w = int(x.shape[1] ** 0.5)
        
        h, w = self.h // p, self.w // p
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, c))
        x = torch.einsum('nhwpqc->nchpwq', x)
        imgs = x.reshape(shape=(x.shape[0], c, h * p, w * p))
        return imgs
    
    def forward(
        self, x, t, y, 
        encoder_mask=None,
        c_size=None,
        ar=None,
        pos_embed=None,
    ):
        """
        Forward pass of DiT.
        x: (N, C, H, W) tensor of spatial inputs (images or latent representations of images)
        t: (N,) tensor of diffusion timesteps
        y: (N, 1, 120, C) tensor of class labels
        """
        if encoder_mask is not None and encoder_mask.ndim == 2:
            encoder_mask = (1 - encoder_mask.to(y.dtype)) * -10000.0
            encoder_mask = encoder_mask.unsqueeze(1)
            
        if pos_embed is None or c_size is None or ar is None:
            kw = get_model_kwargs(x, self)
            c_size, ar, pos_embed = kw["c_size"], kw["ar"], kw["pos_embed"]
        
        input_dtype = x.dtype
        bsz, _, self.h, self.w = x.shape
        
        x = self.x_embedder(x, pos_embed=pos_embed)  # (N, T, D), where T = H * W / patch_size ** 2
        t = self.t_embedder(t, dtype=input_dtype)  # (N, D)
        csize = self.csize_embedder(c_size, bsz, dtype=input_dtype)  # (N, D)
        ar = self.ar_embedder(ar, bsz, dtype=input_dtype)  # (N, D)
        
        temb = t + torch.cat([csize, ar], dim=1)
        t = self.t_block(temb)  # (N, 6, D) for adaLN-single
        y = self.y_embedder(y)  # (N, 1, L, D)

        for block in self.blocks:
            if self.training:
                x = ckpt_wrapper(block, x, t, y, encoder_mask)  # (N, T, D)
            else:
                x = block(x, t, y, encoder_mask)
                          
        x = self.final_layer(x, temb)  # (N, T, patch_size ** 2 * out_channels)
        x = self.unpatchify(x)  # (N, out_channels, H, W)
        return x

# Sine/Cosine Positional Embedding Functions 
def get_2d_sincos_pos_embed(
    embed_dim, grid_size, cls_token=False, extra_tokens=0, interpolation_scale=1.0, base_size=16
):
    """
    grid_size: int of the grid height and width return: pos_embed: [grid_size*grid_size, embed_dim] or
    [1+grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    if isinstance(grid_size, int):
        grid_size = (grid_size, grid_size)

    grid_h = np.arange(grid_size[0], dtype=np.float32) / (grid_size[0] / base_size) / interpolation_scale
    grid_w = np.arange(grid_size[1], dtype=np.float32) / (grid_size[1] / base_size) / interpolation_scale
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size[1], grid_size[0]])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token and extra_tokens > 0:
        pos_embed = np.concatenate([np.zeros([extra_tokens, embed_dim]), pos_embed], axis=0)
    return pos_embed


def get_2d_sincos_pos_embed_from_grid(embed_dim, grid):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[0])  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[1])  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1) # (H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float64)
    omega /= embed_dim / 2.
    omega = 1. / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum('m,d->md', pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out) # (M, D/2)
    emb_cos = np.cos(out) # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb

@functools.cache
def get_cached_micro_condition(h, w):
    c_size = torch.tensor([8 * h, 8 * w])
    ar = torch.tensor([c_size[0] / c_size[1]]) # aspect ratio
    c_size = c_size[None]
    return c_size, ar

@torch.no_grad()
def get_model_kwargs(latents, model):
    h, w = latents.shape[-2:]
    c_size, ar = get_cached_micro_condition(h, w)
    pos_embed = model.x_embedder.get_pos_embed(h, w)
    return {
        "c_size": c_size.clone().to(latents.device),
        "ar": ar.clone().to(latents.device),
        "pos_embed": pos_embed.clone().to(latents.device),
    }

#################################################################################
#                                   DiT Configs                                  #
#################################################################################

def DiT_XL_2(**kwargs):
    print(f"Building DiT-XL-2 model with xformers: {XFORMERS_AVAILABLE}")
    model = DiTModel(depth=28, hidden_size=1152, patch_size=2, num_heads=16, **kwargs)
    return model

@torch.no_grad()    
def sample(model, vae, prompt, tokenizer, text_encoder, negative_prompt, size=(1024,1024), steps=20, guidance_scale=7.5, generator=None):
    
    assert type(prompt) == list, "Prompt must be a list of strings"
    if len(negative_prompt) != len(prompt):
        negative_prompt = [negative_prompt] * len(prompt)
        
    prompt = [prompt.replace("_", " ").lower() for prompt in prompt]
    negative_prompt = [negative_prompt.replace("_", " ").lower() for negative_prompt in negative_prompt]
    text_inputs = tokenizer(
        prompt,
        padding="max_length",
        max_length=120,
        truncation=True,
        add_special_tokens=True,
        return_tensors="pt",
    ) 
    text_input_ids = text_inputs.input_ids
    prompt_attention_mask = text_inputs.attention_mask
    prompt_attention_mask = prompt_attention_mask.cuda()
    prompt_embeds = text_encoder(text_input_ids.cuda(), attention_mask=prompt_attention_mask)
    prompt_embeds = prompt_embeds[0]
    
    uncond_input = tokenizer(
        negative_prompt,
        padding="max_length",
        max_length=120,
        truncation=True,
        return_attention_mask=True,
        add_special_tokens=True,
        return_tensors="pt",
    )
    negative_prompt_attention_mask = uncond_input.attention_mask
    negative_prompt_attention_mask = negative_prompt_attention_mask.cuda()
    negative_prompt_embeds = text_encoder(uncond_input.input_ids.cuda(), attention_mask=negative_prompt_attention_mask)
    negative_prompt_embeds = negative_prompt_embeds[0]
    
    scheduler = DPMSolverMultistepScheduler()
    model_dtype = next(model.parameters()).dtype
    vae.to(model_dtype)
    
    latents_shape = (len(prompt), 4, size[0] // 8, size[1] // 8)
    latents = torch.randn(latents_shape, generator=generator, dtype=torch.float16).cuda()
    latents = latents * scheduler.init_noise_sigma
        
    scheduler.set_timesteps(steps)
    timesteps = scheduler.timesteps
    num_latent_input = 2
    
    cond = torch.cat([negative_prompt_embeds, prompt_embeds]).cuda().to(model_dtype)
    encoder_mask = torch.cat([negative_prompt_attention_mask, prompt_attention_mask]).cuda().to(model_dtype)
    extra_kwargs = {
        "y": cond, 
        "encoder_mask": encoder_mask,
        **get_model_kwargs(latents, model),
    }
            
    # micro-condition
    for i, t in tqdm(enumerate(timesteps), total=steps, desc="Sampling"):
        # expand the latents if we are doing classifier free guidance
        latent_model_input = torch.cat([latents] * 2)
        latent_model_input = scheduler.scale_model_input(latent_model_input, t)

        noise_pred = model(
            x=latent_model_input, 
            t=torch.asarray([t] * latent_model_input.shape[0]).cuda(), 
            **extra_kwargs
        )
        noise_pred_uncond, noise_pred_text = noise_pred.chunk(num_latent_input)  # uncond by negative prompt
        noise_pred = noise_pred_uncond + guidance_scale * (noise_pred_text - noise_pred_uncond)
        
        # compute the previous noisy sample x_t -> x_t-1
        noise_pred  = noise_pred.chunk(2, dim=1)[0]
        latents = scheduler.step(noise_pred, t, latents, return_dict=False)[0]
        
    decoded = vae.decode(latents / vae.config.scaling_factor).sample
    image = torch.clamp((decoded + 1.0) / 2.0, min=0.0, max=1.0).cpu().float()
    image = image.cpu().permute(0, 2, 3, 1).float().numpy()

    image = (image * 255).round().astype("uint8")
    image = [Image.fromarray(im) for im in image]
    return image


ASPECT_RATIO_1024 = {
    '0.25': [512., 2048.], '0.26': [512., 1984.], '0.27': [512., 1920.], '0.28': [512., 1856.],
    '0.32': [576., 1792.], '0.33': [576., 1728.], '0.35': [576., 1664.], '0.4':  [640., 1600.],
    '0.42':  [640., 1536.], '0.48': [704., 1472.], '0.5': [704., 1408.], '0.52': [704., 1344.],
    '0.57': [768., 1344.], '0.6': [768., 1280.], '0.68': [832., 1216.], '0.72': [832., 1152.],
    '0.78': [896., 1152.], '0.82': [896., 1088.], '0.88': [960., 1088.], '0.94': [960., 1024.],
    '1.0':  [1024., 1024.], '1.07': [1024.,  960.], '1.13': [1088.,  960.], '1.21': [1088.,  896.],
    '1.29': [1152.,  896.], '1.38': [1152.,  832.], '1.46': [1216.,  832.], '1.67': [1280.,  768.],
    '1.75': [1344.,  768.], '2.0':  [1408.,  704.], '2.09':  [1472.,  704.], '2.4':  [1536.,  640.],
    '2.5':  [1600.,  640.], '2.89':  [1664.,  576.], '3.0':  [1728.,  576.], '3.11':  [1792.,  576.],
    '3.62':  [1856.,  512.], '3.75':  [1920.,  512.], '3.88':  [1984.,  512.], '4.0':  [2048.,  512.],
}

if __name__ == '__main__':    
    # torch.backends.cuda.matmul.allow_tf32 = True
    # torch.backends.cudnn.allow_tf32 = True
    # torch.set_float32_matmul_precision('medium')
    
    from PIL import Image
    from transformers import T5EncoderModel, T5Tokenizer
    from diffusers import DPMSolverMultistepScheduler, AutoencoderKL

    seed = int(184371982347)
    generator = torch.Generator().manual_seed(seed)
    
    text_encoder = T5EncoderModel.from_pretrained(
        "PixArt-alpha/PixArt-XL-2-1024-MS", 
        torch_dtype=torch.bfloat16, 
        device_map="auto",
        use_safetensors=True, 
        subfolder="text_encoder"
    )
    tokenizer = T5Tokenizer.from_pretrained("PixArt-alpha/PixArt-XL-2-1024-MS", legacy=False, subfolder="tokenizer")
    vae = AutoencoderKL.from_pretrained("stabilityai/sd-vae-ft-mse").cuda()
    
    model = DiT_XL_2(input_size=1024//8, interpolation_scale=2.)
    model.to("cuda:0", dtype=torch.float16, memory_format=torch.channels_last).eval()
    state_dict = torch.hub.load_state_dict_from_url("https://huggingface.co/datasets/nyanko7/tmp-public/resolve/main/PixArt-XL-2-1024-MS.pth", map_location="cuda:0")["state_dict"]            
    
    print("Loading weights from DiT-XL-2-1024-MS.pth")
    result = model.load_state_dict(state_dict, strict=False)
    assert len(result.unexpected_keys) <= 1, f"Unexpected keys: {result.unexpected_keys}, Missing keys: {result.missing_keys}"
    
    # model = torch.compile(model, mode="max-autotune", dynamic=True)
    img = sample(
        model=model, 
        vae=vae, 
        prompt=["flower, rose, solo, blue_flower, blue_rose, blue_eyes, long_hair, looking_at_viewer, petals, sitting, white_hair, bangs, bow, hair_bow, dress, blush, long_sleeves, sailor_collar, water, white_sailor_collar, blunt_bangs, window"],
        negative_prompt="", 
        tokenizer=tokenizer, 
        text_encoder=text_encoder, 
        generator=generator,
        size=(1024, 1024), 
        steps=16, 
        guidance_scale=5
    )
    img[0].save("sample.png")


import torch
import math
from transformers import CLIPTokenizer, CLIPTextModel, CLIPTextModelWithProjection
from models.sgm import UNetModel, AutoencoderKL


def disabled_train(self, mode=True):
    """Overwrite model.train with this function to make sure train/eval mode
    does not change anymore."""
    return self


class AutoencoderKLWrapper(AutoencoderKL):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnetWrapper(torch.nn.Module):
    def __init__(self, config):
        super().__init__()
        self.diffusion_model = UNetModel(**config)

    def forward(
        self, x: torch.Tensor, t: torch.Tensor, c: dict, **kwargs
    ) -> torch.Tensor:
        x = torch.cat((x, c.get("concat", torch.Tensor([]).type_as(x))), dim=1)
        return self.diffusion_model(
            x,
            timesteps=t,
            context=c.get("crossattn", None),
            y=c.get("vector", None),
            **kwargs,
        )


def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.
    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = torch.exp(
        -math.log(max_period)
        * torch.arange(start=0, end=half, dtype=torch.float32)
        / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    if dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    return embedding


def get_timestep_embedding(x, outdim):
    assert len(x.shape) == 2
    b, dims = x.shape[0], x.shape[1]
    x = torch.flatten(x)
    emb = timestep_embedding(x, outdim)
    emb = torch.reshape(emb, (b, dims * outdim))
    return emb


def get_size_embeddings(orig_size, crop_size, target_size, device):
    emb1 = get_timestep_embedding(orig_size, 256)
    emb2 = get_timestep_embedding(crop_size, 256)
    emb3 = get_timestep_embedding(target_size, 256)
    vector = torch.cat([emb1, emb2, emb3], dim=1).to(device)
    return vector


def convert_sdxl_text_encoder_2_checkpoint(checkpoint, max_length=77):
    prefix = "conditioner.embedders.1.model."
    conv_mapping = {
        ".positional_embedding": ".embeddings.position_embedding.weight",
        "text_model.text_projection": "text_projection.weight",
        ".token_embedding.weight": ".embeddings.token_embedding.weight",
        ".ln_final": ".final_layer_norm",
    }

    def convert_key(key):
        key = key.replace(prefix + "transformer.", "text_model.encoder.")
        key = key.replace(prefix, "text_model.")
        if (
            "attn.in_proj" in key
            or "logit_scale" in key
            or "embeddings.position_ids" in key
        ):
            return None
        if ".resblocks." in key:
            key = key.replace(".resblocks.", ".layers.")
            if ".ln_" in key:
                key = key.replace(".ln_", ".layer_norm")
            elif ".mlp." in key:
                key = key.replace(".c_fc.", ".fc1.").replace(".c_proj.", ".fc2.")
            elif ".attn.out_proj" in key:
                key = key.replace(".attn.out_proj.", ".self_attn.out_proj.")
            elif ".attn.in_proj" in key:
                return None
            else:
                raise ValueError(f"unexpected key in SD: {key}")
        else:
            for old, new in conv_mapping.items():
                key = key.replace(old, new)
        return key

    new_sd = {convert_key(k): v for k, v in checkpoint.items() if convert_key(k)}
    for key in checkpoint.keys():
        if ".resblocks" in key and ".attn.in_proj_" in key:
            values = torch.chunk(checkpoint[key], 3)

            key_suffix = ".weight" if "weight" in key else ".bias"
            key_pfx = key.replace(
                prefix + "transformer.resblocks.", "text_model.encoder.layers."
            )
            key_pfx = key_pfx.replace("_weight", "")
            key_pfx = key_pfx.replace("_bias", "")
            key_pfx = key_pfx.replace(".attn.in_proj", ".self_attn.")
            new_sd[key_pfx + "q_proj" + key_suffix] = values[0]
            new_sd[key_pfx + "k_proj" + key_suffix] = values[1]
            new_sd[key_pfx + "v_proj" + key_suffix] = values[2]

    return new_sd


def process_input_ids(input_ids, tokenizer, max_length):
    if max_length > tokenizer.model_max_length:
        input_ids = input_ids.squeeze(0)
        iids_list = []
        for i in range(
            1,
            max_length - tokenizer.model_max_length + 2,
            tokenizer.model_max_length - 2,
        ):
            ids_chunk = (
                input_ids[0].unsqueeze(0),  # BOS
                input_ids[i : i + tokenizer.model_max_length - 2],
                input_ids[-1].unsqueeze(0),  # PAD or EOS
            )
            ids_chunk = torch.cat(ids_chunk)

            if ids_chunk[-2] not in [tokenizer.eos_token_id, tokenizer.pad_token_id]:
                ids_chunk[-1] = tokenizer.eos_token_id
            if ids_chunk[1] == tokenizer.pad_token_id:
                ids_chunk[1] = tokenizer.eos_token_id

            iids_list.append(ids_chunk)

        input_ids = torch.stack(iids_list)
    return input_ids


def get_hidden_states_sdxl(
    prompt: str,
    max_token_length: int,
    tokenizer1: CLIPTokenizer,
    tokenizer2: CLIPTokenizer,
    text_encoder1: CLIPTextModel,
    text_encoder2: CLIPTextModelWithProjection,
):
    device = next(text_encoder1.parameters()).device
    input_ids1 = tokenizer1(
        prompt,
        truncation=True,
        max_length=max_token_length,
        return_overflowing_tokens=False,
        padding="max_length",
        return_tensors="pt",
    ).input_ids
    input_ids2 = tokenizer2(
        prompt,
        truncation=True,
        max_length=max_token_length,
        return_overflowing_tokens=False,
        padding="max_length",
        return_tensors="pt",
    ).input_ids

    input_ids1 = torch.stack(
        [process_input_ids(inp, tokenizer1, max_token_length) for inp in input_ids1]
    ).to(device)
    input_ids2 = torch.stack(
        [process_input_ids(inp, tokenizer2, max_token_length) for inp in input_ids2]
    ).to(device)

    # input_ids: b,n,77 -> b*n, 77
    b_size = input_ids1.size()[0]
    input_ids1 = input_ids1.reshape(
        (-1, tokenizer1.model_max_length)
    )  # batch_size*n, 77
    input_ids2 = input_ids2.reshape(
        (-1, tokenizer2.model_max_length)
    )  # batch_size*n, 77

    # text_encoder1
    enc_out = text_encoder1(input_ids1, output_hidden_states=True, return_dict=True)
    hidden_states1 = enc_out["hidden_states"][11]

    # text_encoder2
    enc_out = text_encoder2(input_ids2, output_hidden_states=True, return_dict=True)
    hidden_states2 = enc_out["hidden_states"][-2]

    pool2 = enc_out["text_embeds"]

    # b*n, 77, 768 or 1280 -> b, n*77, 768 or 1280
    n_size = 1 if max_token_length is None else max_token_length // 75
    hidden_states1 = hidden_states1.reshape((b_size, -1, hidden_states1.shape[-1]))
    hidden_states2 = hidden_states2.reshape((b_size, -1, hidden_states2.shape[-1]))

    if max_token_length is not None:
        states_list = [hidden_states1[:, 0].unsqueeze(1)]  # <BOS>
        for i in range(1, max_token_length, tokenizer1.model_max_length):
            states_list.append(
                hidden_states1[:, i : i + tokenizer1.model_max_length - 2]
            )
        states_list.append(hidden_states1[:, -1].unsqueeze(1))  # <EOS>
        hidden_states1 = torch.cat(states_list, dim=1)

        states_list = [hidden_states2[:, 0].unsqueeze(1)]  # <BOS>
        for i in range(1, max_token_length, tokenizer2.model_max_length):
            chunk = hidden_states2[:, i : i + tokenizer2.model_max_length - 2]
            states_list.append(chunk)

        states_list.append(hidden_states2[:, -1].unsqueeze(1))
        hidden_states2 = torch.cat(states_list, dim=1)
        pool2 = pool2[::n_size]

    return hidden_states1, hidden_states2, pool2



import torch
import torch.utils.checkpoint
import lightning as pl
from PIL import Image
from omegaconf import OmegaConf

import torch

from models.sgm import GeneralConditioner
from modules.sdxl_utils import disabled_train, UnetWrapper, AutoencoderKLWrapper
from common.utils import load_torch_file, rank_zero_print, EmptyInitWrapper

from diffusers import EulerDiscreteScheduler, DDPMScheduler
from lightning.pytorch.utilities import rank_zero_only
from safetensors.torch import save_file


# define the LightningModule
class StableDiffusionModel(pl.LightningModule):
    def __init__(self, model_path, config, device):
        super().__init__()
        self.config = config
        self.model_path = model_path
        self.target_device = device
        self.init_model()

    def build_models(self, init_unet=True, init_vae=True, init_conditioner=True):
        trainer_cfg = self.config.trainer
        config = self.config
        advanced = config.get("advanced", {})

        model_config = OmegaConf.load("modules/sdxl_base.yaml")
        model_params = model_config.model.params

        if trainer_cfg.use_xformers:
            unet_config = model_params.network_config.params
            vae_config = model_params.first_stage_config.params
            unet_config.spatial_transformer_attn_type = "softmax-xformers"
            vae_config.ddconfig.attn_type = "vanilla-xformers"

        for conditioner in model_params.conditioner_config.params.emb_models:
            if "CLIPEmbedder" not in conditioner.target:
                continue
            self.max_token_length = self.config.dataset.get("max_token_length", 75) + 2
            conditioner.params["device"] = str(self.target_device)
            conditioner.params["max_length"] = self.max_token_length

        tte_1 = advanced.get("train_text_encoder_1", False)
        tte_2 = advanced.get("train_text_encoder_2", False)
        model_params.conditioner_config.params.emb_models[0]["is_trainable"] = tte_1
        model_params.conditioner_config.params.emb_models[1]["is_trainable"] = tte_2

        self.scale_factor = model_params.scale_factor
        with EmptyInitWrapper(self.target_device):
            vae_config = model_params.first_stage_config.params
            unet_config = model_params.network_config.params
            cond_config = model_params.conditioner_config.params

            vae = AutoencoderKLWrapper(**vae_config) if init_vae else None
            unet = UnetWrapper(unet_config) if init_unet else None
            conditioner = (
                GeneralConditioner(**cond_config) if init_conditioner else None
            )

        vae.train = disabled_train
        vae.eval()
        vae.requires_grad_(False)
        return vae, unet, conditioner

    def init_model(self):
        sd = load_torch_file(self.model_path, self.target_device)
        self.first_stage_model, self.model, self.conditioner = self.build_models()
        self.noise_scheduler = DDPMScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            num_train_timesteps=1000,
            clip_sample=False,
        )
        self.to(self.target_device)

        rank_zero_print(f"Loading model from {self.model_path}")
        missing, unexpected = self.load_state_dict(sd, strict=False)
        if len(missing) > 0:
            rank_zero_print(f"Missing Keys: {missing}")
        if len(unexpected) > 0:
            rank_zero_print(f"Unexpected Keys: {unexpected}")
            
        self.batch_size = self.config.trainer.batch_size
        self.vae_encode_bsz = self.config.get("vae_encode_batch_size", self.batch_size)
        if self.vae_encode_bsz < 0:
            self.vae_encode_bsz = self.batch_size
            
    def encode_pixels(self, inputs):
        feed_pixel_values = inputs
        latents = []
        for i in range(0, feed_pixel_values.shape[0], self.vae_encode_bsz):
            latents.append(
                self.vae.encode(feed_pixel_values[i : i + self.vae_encode_bsz]).latent_dist.sample()
            )
        latents = torch.cat(latents, dim=0)
        latents = latents * self.vae.config.scaling_factor
        return latents

    @torch.no_grad()
    def decode_first_stage(self, z):
        z = 1.0 / self.scale_factor * z
        with torch.autocast("cuda", enabled=False):
            out = self.first_stage_model.decode(z)
        return out

    @torch.no_grad() 
    def encode_first_stage(self, x):
        latents = []
        with torch.autocast("cuda", enabled=False):
            for i in range(0, x.shape[0], self.vae_encode_bsz):
                latents.append(self.first_stage_model.encode(x).sample())
        z = torch.cat(latents, dim=0)
        z = self.scale_factor * z
        return z

    @torch.inference_mode()
    @rank_zero_only
    def sample(
        self,
        prompt,
        negative_prompt="lowres, low quality, text, error, extra digit, cropped",
        generator=None,
        size=(1024, 1024),
        steps=20,
        guidance_scale=6.5,
    ):
        self.model.eval()
        self.first_stage_model.to(self.target_device)

        model_dtype = next(self.model.parameters()).dtype
        scheduler = EulerDiscreteScheduler(
            beta_start=0.00085,
            beta_end=0.012,
            beta_schedule="scaled_linear",
            num_train_timesteps=1000,
        )
        prompts_batch = {
            "target_size_as_tuple": torch.stack([torch.asarray(size)]).cuda(),
            "original_size_as_tuple": torch.stack([torch.asarray(size)]).cuda(),
            "crop_coords_top_left": torch.stack([torch.asarray((0, 0))]).cuda(),
        }
        prompts_batch["prompts"] = prompt
        cond = self.encode_batch(prompts_batch)
        prompts_batch["prompts"] = negative_prompt
        uncond = self.encode_batch(prompts_batch)

        crossattn = torch.cat([uncond["crossattn"], cond["crossattn"]], dim=0)
        vector = torch.cat([uncond["vector"], cond["vector"]], dim=0)
        cond = {
            "crossattn": crossattn.cuda().to(model_dtype),
            "vector": vector.cuda().to(model_dtype),
        }

        latents_shape = (1, 4, size[0] // 8, size[1] // 8)
        latents = torch.randn(latents_shape, generator=generator, dtype=torch.float32)
        latents = latents * scheduler.init_noise_sigma

        scheduler.set_timesteps(steps)
        timesteps = scheduler.timesteps
        num_latent_input = 2
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = latents.repeat((num_latent_input, 1, 1, 1))
            latent_model_input = scheduler.scale_model_input(latent_model_input, t)
            latent_model_input = latent_model_input.cuda().to(model_dtype)

            noise_pred = self.model(latent_model_input, torch.asarray([t]).cuda(), cond)
            pred_uncond, pred_text = noise_pred.chunk(
                num_latent_input
            )  # uncond by negative prompt
            noise_pred = pred_uncond + guidance_scale * (pred_text - pred_uncond)

            # compute the previous noisy sample x_t -> x_t-1
            latents = scheduler.step(noise_pred, t, latents.cuda()).prev_sample

        latents = self.decode_first_stage(latents.to(torch.float32))
        image = torch.clamp((latents + 1.0) / 2.0, min=0.0, max=1.0).cpu().float()
        image = image.cpu().permute(0, 2, 3, 1).float().numpy()

        image = (image * 255).round().astype("uint8")
        image = [Image.fromarray(im) for im in image]

        self.model.train()
        return image

    def save_checkpoint(self, model_path):
        cfg = self.config.trainer
        string_cfg = OmegaConf.to_yaml(self.config)
        if cfg.get("save_format") == "safetensors":

            model_path += ".safetensors"
            state_dict = self.model.state_dict()
            # check if any keys startswith modules. if so, remove the modules. prefix
            if any([key.startswith("module.") for key in state_dict.keys()]):
                state_dict = {
                    key.replace("module.", ""): value
                    for key, value in state_dict.items()
                }
            save_file(state_dict, model_path, metadata={"trainer_config": string_cfg})
        else:
            model_path += ".ckpt"
            torch.save(
                model_path,
                {"state_dict": self.model.state_dict(), "trainer_config": string_cfg},
            )
        rank_zero_print(f"Saved model to {model_path}")

    def forward(self, batch):
        raise NotImplementedError



import torch
import os
import lightning as pl

from omegaconf import OmegaConf
from PIL import Image
from safetensors.torch import save_file
from transformers import T5EncoderModel, T5Tokenizer
from diffusers import AutoencoderKL, DPMSolverMultistepScheduler
from lightning.pytorch.utilities.model_summary import ModelSummary

from common.utils import load_torch_file, rank_zero_print, get_class
from common.dataset import AspectRatioDataset, worker_init_fn

from models.pixart.dit import DiT_XL_2, get_model_kwargs
from models.pixart.diffusion import SpacedDiffusion, get_named_beta_schedule, space_timesteps
from models.pixart.diffusion import ModelVarType, ModelMeanType, LossType


def setup(fabric: pl.Fabric, config: OmegaConf) -> tuple:
    model_path = config.trainer.model_path
    model = DiffusionModel(
        model_path=model_path, 
        config=config, 
        device=fabric.device
    )
    dataset = AspectRatioDataset(
        batch_size=config.trainer.batch_size,
        rank=fabric.global_rank,
        dtype=torch.float32,
        base_len=config.trainer.resolution,
        **config.dataset,
    )
    dataloader = torch.utils.data.DataLoader(
        dataset,
        sampler=None,
        batch_size=None,
        persistent_workers=False,
        num_workers=config.dataset.get("num_workers", 4),
        worker_init_fn=worker_init_fn,
        shuffle=False,
        pin_memory=True,
    )
    
    params_to_optim = [{'params': model.model.parameters()}]
    optim_param = config.optimizer.params
    optimizer = get_class(config.optimizer.name)(
        params_to_optim, **optim_param
    )
    scheduler = None
    if config.get("scheduler"):
        scheduler = get_class(config.scheduler.name)(
            optimizer, **config.scheduler.params
        )
        
    if fabric.is_global_zero and os.name != "nt":
        print(f"\n{ModelSummary(model, max_depth=1)}\n")
        
    model, optimizer = fabric.setup(model, optimizer)
    dataloader = fabric.setup_dataloaders(dataloader)
    return model, dataset, dataloader, optimizer, scheduler


class DiffusionModel(pl.LightningModule):
    def __init__(self, model_path, config, device):
        super().__init__()
        self.config = config
        self.model_path = model_path
        self.target_device = device
        self.init_model()

    def init_model(self):
        t5_model_path = self.config.trainer.get("t5_model_path", "PixArt-alpha/PixArt-XL-2-1024-MS")
        vae_model_path = self.config.trainer.get("vae_model_path", "stabilityai/sd-vae-ft-mse")
        self.tokenizer = T5Tokenizer.from_pretrained(
           t5_model_path, 
           legacy=False, 
           subfolder="tokenizer"
        )
        self.text_encoder = T5EncoderModel.from_pretrained(
            t5_model_path, 
            torch_dtype=torch.bfloat16,
            use_safetensors=True, 
            subfolder="text_encoder"
        )
        self.vae = AutoencoderKL.from_pretrained(vae_model_path)
        self.text_encoder.requires_grad_(False)
        self.vae.requires_grad_(False)
        
        dit_state_dict = load_torch_file(self.model_path)
        betas = get_named_beta_schedule(
            schedule_name="linear", 
            num_diffusion_timesteps=1000
        )
        timesteps = space_timesteps(
            num_timesteps=1000,
            section_counts=[1000]
        )
        self.diffusion = SpacedDiffusion(
            use_timesteps=timesteps,
            betas=betas,
            model_mean_type=ModelMeanType.EPSILON,
            model_var_type=ModelVarType.LEARNED_RANGE,
            loss_type=LossType.MSE,
            snr=False,
            return_startx=False,
            # rescale_timesteps=rescale_timesteps,
        )
        
        base = self.config.trainer.resolution
        self.model = DiT_XL_2(input_size=base//8, interpolation_scale=base//512)
        self.model.to(memory_format=torch.channels_last).train()
    
        rank_zero_print("Loading weights from checkpoint: DiT-XL-2-1024-MS.pth")
        result = self.model.load_state_dict(dit_state_dict, strict=False)
        assert result.unexpected_keys == ['pos_embed'], f"Unexpected keys: {result.unexpected_keys}, Missing keys: {result.missing_keys}"

    @torch.no_grad()
    def encode_tokens(self, prompts):
        with torch.autocast("cuda", enabled=False):
            text_encoder = self.text_encoder
            text_inputs = self.tokenizer(
                prompts,
                padding="max_length",
                max_length=120,
                truncation=True,
                add_special_tokens=True,
                return_tensors="pt",
            )
            text_input_ids = text_inputs.input_ids.to(text_encoder.device)
            prompt_attention_mask = text_inputs.attention_mask.to(text_encoder.device)

            prompt_embeds = text_encoder(
                text_input_ids, attention_mask=prompt_attention_mask
            )[0]
            return prompt_embeds, prompt_attention_mask

    def forward(self, batch):

        prompts = batch["prompts"]
        prompt_embeds, prompt_attention_mask = self.encode_tokens(prompts)
        
        if not batch["is_latent"]:
            self.vae.to(self.target_device)
            latent_dist = self.vae.encode(batch["pixels"]).latent_dist
            latents = latent_dist.sample() * self.vae.config.scaling_factor
            if torch.any(torch.isnan(latents)):
                rank_zero_print("NaN found in latents, replacing with zeros")
                latents = torch.where(torch.isnan(latents), torch.zeros_like(latents), latents)
        else:
            self.vae.cpu()
            latents = batch["pixels"]
            
        model_dtype = next(self.model.parameters()).dtype
        bsz = latents.shape[0]
        
        # Forward pass through the model
        timesteps = torch.randint(0, 1000, (bsz,), device=latents.device).long()
        latents = latents.to(model_dtype)
        model_kwg = get_model_kwargs(latents, self.model)
        loss = self.diffusion.training_losses(
            self.model, latents, timesteps, 
            model_kwargs=dict(
                y=prompt_embeds.to(self.target_device, dtype=model_dtype), 
                encoder_mask=prompt_attention_mask.to(self.target_device, dtype=model_dtype),
                **model_kwg,
            )
        )['loss'].mean()
        
        if torch.isnan(loss).any() or torch.isinf(loss).any():
            raise FloatingPointError("Error infinite or NaN loss detected")
        
        return loss

    @torch.no_grad()
    def sample(
        self, 
        prompt="1girl, solo", 
        negative_prompt="lowres, low quality, text, error, extra digit, cropped",
        size=(1152, 832),
        steps = 28,
        guidance_scale=6.5,
    ):
        self.model.eval()
        
        prompt = prompt.replace("_", " ").replace("[", "").replace("]", "").replace("(", "").replace(")", "").replace("{", "").replace("}", "").lower()
        scheduler = DPMSolverMultistepScheduler()
        prompt_embeds, prompt_attention_mask = self.encode_tokens(
            [negative_prompt, prompt]
        )

        latents = torch.randn(1, 4, size[0] // 8, size[1] // 8, device=self.target_device)
        scheduler.set_timesteps(steps)
        timesteps = scheduler.timesteps
        latents = latents * scheduler.init_noise_sigma
        
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat([latents] * 2)
            latent_model_input = scheduler.scale_model_input(latent_model_input, t)

            noise_pred = self.model(
                x=latent_model_input, 
                t=torch.stack([t] * latents.shape[0]).to(latents.device),
                y=prompt_embeds.to(latents.device, dtype=latents.dtype), 
                encoder_mask=prompt_attention_mask.to(latents.device, dtype=latents.dtype),
                **get_model_kwargs(latents, self.model),
            )
            noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)  # uncond by negative prompt
            noise_pred = noise_pred_uncond + guidance_scale * (noise_pred_text - noise_pred_uncond)

            # compute the previous noisy sample x_t -> x_t-1
            noise_pred = noise_pred.chunk(2, dim=1)[0]
            latents = scheduler.step(noise_pred, t, latents, return_dict=False)[0]

        decoded = self.vae.decode(latents / self.vae.config.scaling_factor).sample
        image = torch.clamp((decoded + 1.0) / 2.0, min=0.0, max=1.0).detach().float()
        image = image.cpu().permute(0, 2, 3, 1).float().numpy()
        
        image = (image * 255).round().astype("uint8")
        image = [Image.fromarray(im) for im in image]
        
        self.model.train()
        return image
    
    def save_checkpoint(self, model_path):
        cfg = self.config.trainer
        string_cfg = OmegaConf.to_yaml(self.config)
        if cfg.get("save_format") == "safetensors":

            model_path += ".safetensors"
            state_dict = self.model.state_dict()
            # check if any keys startswith modules. if so, remove the modules. prefix
            if any([key.startswith("module.") for key in state_dict.keys()]):
                state_dict = {
                    key.replace("module.", ""): value
                    for key, value in state_dict.items()
                }
            save_file(state_dict, model_path, metadata={"trainer_config": string_cfg})
        else:
            model_path += ".ckpt"
            torch.save(
                model_path,
                {"state_dict": self.model.state_dict(), "trainer_config": string_cfg},
            )
        rank_zero_print(f"Saved model to {model_path}")


from setuptools import setup, find_packages

setup(
    name='naifu',
    version='0.1.3',
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    py_modules=['trainer'],
    description='naifu is designed for training generative models with various configurations and features.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=open('requirements.txt').read().splitlines(),
    url='https://github.com/mikubill/naifu-diffusion',
    entry_points={
        'console_scripts': [
            'naifu = trainer:main',
        ],
    },
    author='Mikubill',
    author_email='31246794+Mikubill@users.noreply.github.com'
)


# python trainer.py --model_path=/tmp/model --config config/test.yaml
import copy
import os, sys

# Hide welcome message from bitsandbytes
os.environ.update({"BITSANDBYTES_NOWELCOME": "1"})

import torch
import lightning as pl
import argparse

from common.utils import *
from omegaconf import OmegaConf
from pathlib import Path
from tqdm import tqdm
from lightning.pytorch.loggers import WandbLogger
from lightning.pytorch.utilities import rank_zero_only

torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.allow_tf32 = True


class Trainer:
    def __init__(self, fabric: pl.Fabric, config: OmegaConf):
        self.fabric = fabric

        model_cls = get_class(config.target)
        model, dataset, dataloader, optimizer, scheduler = model_cls(fabric, config)

        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.dataset = dataset
        self.dataloader = dataloader

    def prepare_logger(self):
        fabric = self.fabric
        if fabric.logger:
            config = OmegaConf.to_container(self.model.config, resolve=True)
            fabric.logger.log_hyperparams(config)
            
    def on_post_training_batch(self, global_step, current_epoch, state):
        config = self.model.config
        fabric = self.fabric

        if fabric.logger:
            self.log_lr_values(global_step, fabric)

        self.perform_sampling(global_step, current_epoch)
        self.save_model(global_step, current_epoch, state)

    def log_lr_values(self, global_step, fabric):
        optimizer_name = self.model.config.optimizer.name
        last_lr = [group.get("lr", 0) for group in self.optimizer.param_groups]
        ocls = self.optimizer.__class__.__name__

        for i, lr in enumerate(last_lr):
            fabric.log(f"lr-{ocls}-{i}", lr, step=global_step)

        is_da = optimizer_name.startswith("DAdapt")
        is_prodigy = optimizer_name.startswith("prodigyopt")
        if not (is_da or is_prodigy):
            return

        last_d_lr = [(g["d"] * g["lr"]) for g in self.optimizer.param_groups]
        for i, lr in enumerate(last_d_lr):
            fabric.log(f"d*lr-{ocls}-{i}", lr, step=global_step)

    def save_model(self, global_step, current_epoch, state, is_last=False):
        config = self.model.config
        cfg = config.trainer
        fabric = self.fabric
        ckpt_st = cfg.checkpoint_steps
        ckpt_fq = cfg.checkpoint_freq
        ckpt_dir = cfg.checkpoint_dir

        is_ckpt_step = ckpt_st > 0 and global_step % ckpt_st == 0
        is_ckpt_epoch = ckpt_fq > 0 and current_epoch % ckpt_fq == 0
        to_save = (is_last and is_ckpt_epoch) or is_ckpt_step

        if not (fabric.is_global_zero and to_save):
            return

        model_path = os.path.join(ckpt_dir, f"nd-checkpoint-e{current_epoch:02d}")
        if is_ckpt_step:
            model_path = os.path.join(ckpt_dir, f"nd-checkpoint-s{global_step:02d}")
            
        self.model.save_checkpoint(model_path)

    def perform_sampling(self, global_step, current_epoch, is_last=False):
        config = self.model.config
        enabled_sampling = self.fabric.is_global_zero and config.sampling.enabled
        if not enabled_sampling:
            return

        sampling_cfg = config.sampling
        sampling_steps = sampling_cfg.every_n_steps
        sample_by_step = sampling_steps > 0 and global_step % sampling_steps == 0
        sampling_epochs = sampling_cfg.every_n_epochs
        sample_by_epoch = sampling_epochs > 0 and current_epoch % sampling_epochs == 0

        if (is_last and sample_by_epoch) or sample_by_step:
            self.sampler(
                logger=self.fabric.logger, 
                config=sampling_cfg, 
                model=self.model, 
                current_epoch=current_epoch, 
                global_step=global_step
            )

    @rank_zero_only
    def sampler(self, logger, config, model, current_epoch, global_step):
        if not any(config.prompts):
            return

        rng_state = torch.get_rng_state()
        cuda_rng_state = torch.cuda.get_rng_state()

        save_dir = Path(config.save_dir)
        save_dir.mkdir(parents=True, exist_ok=True)
        generator = torch.Generator(device="cpu").manual_seed(config.seed)
        prompts = list(config.prompts)
        prompt_to_gen = copy.deepcopy(prompts)
        images = []
        for prompt in zip(prompt_to_gen):
            images.extend(model.sample(prompt, generator=generator))

        for j, image in enumerate(images):
            image.save(save_dir / f"nd_sample_e{current_epoch}_s{global_step}_{j}.png")

        if config.use_wandb and logger:
            logger.log_image(
                key="samples", 
                images=images, 
                caption=prompts, 
                step=global_step
            )

        torch.cuda.empty_cache()
        torch.set_rng_state(rng_state)
        torch.cuda.set_rng_state(cuda_rng_state)

    def train(self):
        config = self.model.config
        cfg = config.trainer
        fabric = self.fabric
        grad_accum_steps = cfg.accumulate_grad_batches
        grad_clip_val = cfg.gradient_clip_val

        local_step = global_step = 0
        current_epoch = 0
        should_stop = False

        state = {"state_dict": self.model}
        os.makedirs(cfg.checkpoint_dir, exist_ok=True)
        
        if not cfg.get("save_weights_only", False):
            state.update({"optimizer": self.optimizer})

        if Path(cfg.checkpoint_dir).is_dir() and cfg.get("resume"):
            latest_checkpoint_path = get_latest_checkpoint(cfg.checkpoint_dir)
            remainder = self.fabric.load(latest_checkpoint_path, state)
            global_step = remainder.pop("global_step")
            current_epoch = remainder.pop("current_epoch")
            rank_zero_print(f"Resuming from checkpoint {latest_checkpoint_path}")

        if cfg.max_epochs > 0 and current_epoch >= cfg.max_epochs:
            should_stop = True

        rank_zero_print(f"Starting training from epoch {current_epoch}")
        self.prepare_logger()
        prog_bar = tqdm(
            self.dataloader,
            total=len(self.dataloader) // grad_accum_steps - 1,
            desc=f"Epoch {current_epoch}",
            disable=not fabric.is_global_zero,
        )

        loss_rec = LossRecorder()
        assert len(self.dataloader) > 0, "Dataloader is empty"
        while not should_stop:
            if fabric.is_global_zero:
                prog_bar.refresh()
                prog_bar.reset()
                prog_bar.total = len(self.dataloader) // grad_accum_steps
                prog_bar.set_description(f"Epoch {current_epoch}")

            for batch_idx, batch in enumerate(self.dataloader):
                local_step += 1
                is_accumulating = local_step % grad_accum_steps != 0

                with fabric.no_backward_sync(self.model, enabled=is_accumulating):
                    loss = self.model(batch)
                    self.fabric.backward(loss / grad_accum_steps)

                loss_rec.add(epoch=current_epoch, step=batch_idx, loss=loss.item())
                postfix = (
                    f"train_loss: {loss:.3f}, avg_loss: {loss_rec.moving_average:.3f}"
                )
                prog_bar.set_postfix_str(postfix)

                if is_accumulating:
                    # skip here if we are accumulating
                    continue

                if grad_clip_val > 0:
                    val = grad_clip_val
                    self.fabric.clip_gradients(self.model, self.optimizer, max_norm=val)

                self.optimizer.step()
                self.optimizer.zero_grad(set_to_none=True)

                if self.scheduler is not None:
                    fp_batch = current_epoch + batch_idx / len(self.dataloader)
                    if "transformers" in config.scheduler.name:
                        self.scheduler.step(global_step)
                    else:
                        self.scheduler.step(fp_batch)

                if fabric.logger:
                    fabric.log("train_loss", loss, step=global_step)

                global_step += 1
                prog_bar.update(1)
                prog_bar.set_postfix_str(postfix)
                state.update(global_step=global_step, current_epoch=current_epoch)
                self.on_post_training_batch(global_step, current_epoch, state)

            current_epoch += 1
            if cfg.max_epochs > 0 and current_epoch >= cfg.max_epochs:
                should_stop = True

            state.update(global_step=global_step, current_epoch=current_epoch)
            self.perform_sampling(global_step, current_epoch, is_last=True)
            self.save_model(global_step, current_epoch, state, is_last=True)
            
            
class NonAutocastMixedPrecision(pl.fabric.plugins.precision.amp.MixedPrecision):
    """Mixed precision training without automatic casting inputs and outputs."""

    def convert_input(self, data):
        return data

    def convert_output(self, data):
        return data


class LossRecorder:
    def __init__(self):
        self.loss_list = []
        self.loss_total = 0.0

    def add(self, *, epoch: int, step: int, loss: float) -> None:
        if epoch == 0 or len(self.loss_list) <= step:
            self.loss_list.append(loss)
        else:
            self.loss_total -= self.loss_list[step]
            self.loss_list[step] = loss
        self.loss_total += loss

    @property
    def moving_average(self) -> float:
        if len(self.loss_list) == 0:
            return 0.0
        return self.loss_total / len(self.loss_list)


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--config', type=str)
    group.add_argument('config', nargs='?', type=str)
    parser.add_argument("--resume", action="store_true")
    args = parser.parse_args()
    
    config = OmegaConf.load(args.config)
    config.trainer.resume = args.resume
    plugins = []
    
    strategy = config.lightning.get("strategy", "auto")
    if "." in strategy:
        strategy = get_class(strategy)

    if os.environ.get("SM_TRAINING", False):
        strategy, config = setup_smddp(config)

    target_precision = config.lightning.precision
    if target_precision in ["16-mixed", "bf16-mixed"]:
        config.lightning.precision = None
        plugins.append(NonAutocastMixedPrecision(target_precision, "cuda"))

    loggers = pl.fabric.loggers.CSVLogger(".")
    if config.trainer.wandb_id != "":
        loggers = WandbLogger(project=config.trainer.wandb_id)
        
    fabric = pl.Fabric(
        loggers=[loggers], 
        plugins=plugins, 
        strategy=strategy, 
        **config.lightning
    )
    fabric.launch()
    fabric.seed_everything(config.trainer.seed)
    Trainer(fabric, config).train()


if __name__ == "__main__":
    main()
